#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from dracclient import constants
from dracclient.resources import uris
from dracclient import utils


class LifecycleControllerManagement(object):

    def __init__(self, client):
        """Creates LifecycleControllerManagement object

        :param client: an instance of WSManClient
        """
        self.client = client

    def get_version(self):
        """Returns the Lifecycle controller version

        :returns: Lifecycle controller version as a tuple of integers
        :raises: WSManRequestFailure on request failures
        :raises: WSManInvalidResponse when receiving invalid response
        :raises: DRACOperationFailed on error reported back by the DRAC
                 interface
        """

        doc = self.client.enumerate(uris.DCIM_SystemView, wait_for_idrac=False)
        lc_version_str = utils.find_xml(doc, 'LifecycleControllerVersion',
                                        uris.DCIM_SystemView).text

        return tuple(map(int, (lc_version_str.split('.'))))


class LCAttribute(object):
    """Generic LC attribute class"""

    def __init__(self, name, instance_id, current_value, pending_value,
                 read_only):
        """Creates LCAttribute object

        :param name: name of the LC attribute
        :param instance_id: InstanceID of the LC attribute
        :param current_value: current value of the LC attribute
        :param pending_value: pending value of the LC attribute, reflecting
                an unprocessed change (eg. config job not completed)
        :param read_only: indicates whether this LC attribute can be changed
        """
        self.name = name
        self.instance_id = instance_id
        self.current_value = current_value
        self.pending_value = pending_value
        self.read_only = read_only

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    @classmethod
    def parse(cls, namespace, lifecycle_attr_xml):
        """Parses XML and creates LCAttribute object"""

        name = utils.get_wsman_resource_attr(
            lifecycle_attr_xml, namespace, 'AttributeName')
        instance_id = utils.get_wsman_resource_attr(
            lifecycle_attr_xml, namespace, 'InstanceID')
        current_value = utils.get_wsman_resource_attr(
            lifecycle_attr_xml, namespace, 'CurrentValue', nullable=True)
        pending_value = utils.get_wsman_resource_attr(
            lifecycle_attr_xml, namespace, 'PendingValue', nullable=True)
        read_only = utils.get_wsman_resource_attr(
            lifecycle_attr_xml, namespace, 'IsReadOnly')

        return cls(name, instance_id, current_value, pending_value,
                   (read_only == 'true'))


class LCEnumerableAttribute(LCAttribute):
    """Enumerable LC attribute class"""

    namespace = uris.DCIM_LCEnumeration

    def __init__(self, name, instance_id, current_value, pending_value,
                 read_only, possible_values):
        """Creates LCEnumerableAttribute object

        :param name: name of the LC attribute
        :param current_value: current value of the LC attribute
        :param pending_value: pending value of the LC attribute, reflecting
                an unprocessed change (eg. config job not completed)
        :param read_only: indicates whether this LC attribute can be changed
        :param possible_values: list containing the allowed values for the LC
                                attribute
        """
        super(LCEnumerableAttribute, self).__init__(name, instance_id,
                                                    current_value,
                                                    pending_value, read_only)
        self.possible_values = possible_values

    @classmethod
    def parse(cls, lifecycle_attr_xml):
        """Parses XML and creates LCEnumerableAttribute object"""

        lifecycle_attr = LCAttribute.parse(cls.namespace, lifecycle_attr_xml)
        possible_values = [attr.text for attr
                           in utils.find_xml(lifecycle_attr_xml,
                                             'PossibleValues',
                                             cls.namespace, find_all=True)]

        return cls(lifecycle_attr.name, lifecycle_attr.instance_id,
                   lifecycle_attr.current_value, lifecycle_attr.pending_value,
                   lifecycle_attr.read_only, possible_values)

    def validate(self, new_value):
        """Validates new value"""

        if str(new_value) not in self.possible_values:
            msg = ("Attribute '%(attr)s' cannot be set to value '%(val)s'."
                   " It must be in %(possible_values)r.") % {
                       'attr': self.name,
                       'val': new_value,
                       'possible_values': self.possible_values}
            return msg


class LCStringAttribute(LCAttribute):
    """String LC attribute class"""

    namespace = uris.DCIM_LCString

    def __init__(self, name, instance_id, current_value, pending_value,
                 read_only, min_length, max_length):
        """Creates LCStringAttribute object

        :param name: name of the LC attribute
        :param instance_id: InstanceID of the LC attribute
        :param current_value: current value of the LC attribute
        :param pending_value: pending value of the LC attribute, reflecting
                an unprocessed change (eg. config job not completed)
        :param read_only: indicates whether this LC attribute can be changed
        :param min_length: minimum length of the string
        :param max_length: maximum length of the string
        """
        super(LCStringAttribute, self).__init__(name, instance_id,
                                                current_value, pending_value,
                                                read_only)
        self.min_length = min_length
        self.max_length = max_length

    @classmethod
    def parse(cls, lifecycle_attr_xml):
        """Parses XML and creates LCStringAttribute object"""

        lifecycle_attr = LCAttribute.parse(cls.namespace, lifecycle_attr_xml)
        min_length = int(utils.get_wsman_resource_attr(
            lifecycle_attr_xml, cls.namespace, 'MinLength'))
        max_length = int(utils.get_wsman_resource_attr(
            lifecycle_attr_xml, cls.namespace, 'MaxLength'))

        return cls(lifecycle_attr.name, lifecycle_attr.instance_id,
                   lifecycle_attr.current_value, lifecycle_attr.pending_value,
                   lifecycle_attr.read_only, min_length, max_length)


class LCConfiguration(object):

    NAMESPACES = [(uris.DCIM_LCEnumeration, LCEnumerableAttribute),
                  (uris.DCIM_LCString, LCStringAttribute)]

    def __init__(self, client):
        """Creates LifecycleControllerManagement object

        :param client: an instance of WSManClient
        """
        self.client = client

    def list_lifecycle_settings(self, by_name=False):
        """List the LC configuration settings

        :param by_name: Controls whether returned dictionary uses Lifecycle
                        attribute name or instance_id as key.
        :returns: a dictionary with the LC settings using InstanceID as the
                  key. The attributes are either LCEnumerableAttribute,
                  LCStringAttribute or LCIntegerAttribute objects.
        :raises: WSManRequestFailure on request failures
        :raises: WSManInvalidResponse when receiving invalid response
        :raises: DRACOperationFailed on error reported back by the DRAC
                 interface
        """
        return utils.list_settings(self.client, self.NAMESPACES, by_name)

    def is_lifecycle_in_recovery(self):
        """Check if Lifecycle Controller in recovery mode or not

        This method checks the LCStatus value to determine if lifecycle
        controller is in recovery mode by invoking GetRemoteServicesAPIStatus
        from iDRAC.

        :returns: a boolean indicating if lifecycle controller is in recovery
        :raises: WSManRequestFailure on request failures
        :raises: WSManInvalidResponse when receiving invalid response
        :raises: DRACOperationFailed on error reported back by the DRAC
                 interface
        """

        selectors = {'SystemCreationClassName': 'DCIM_ComputerSystem',
                     'SystemName': 'DCIM:ComputerSystem',
                     'CreationClassName': 'DCIM_LCService',
                     'Name': 'DCIM:LCService'}

        doc = self.client.invoke(uris.DCIM_LCService,
                                 'GetRemoteServicesAPIStatus',
                                 selectors,
                                 {},
                                 expected_return_value=utils.RET_SUCCESS,
                                 wait_for_idrac=False)

        lc_status = utils.find_xml(doc,
                                   'LCStatus',
                                   uris.DCIM_LCService).text

        return lc_status == constants.LC_IN_RECOVERY

    def set_lifecycle_settings(self, settings):
        """Sets the Lifecycle Controller configuration

        It sets the pending_value parameter for each of the attributes
        passed in. For the values to be applied, a config job must
        be created.

        :param settings: a dictionary containing the proposed values, with
                         each key being the name of attribute and the value
                         being the proposed value.
        :returns: a dictionary containing:
                 - The is_commit_required key with a boolean value indicating
                   whether a config job must be created for the values to be
                   applied.
                 - The is_reboot_required key with a RebootRequired enumerated
                   value indicating whether the server must be rebooted for the
                   values to be applied.  Possible values are true and false.
        :raises: WSManRequestFailure on request failures
        :raises: WSManInvalidResponse when receiving invalid response
        :raises: DRACOperationFailed on error reported back by the DRAC
                 interface
        """

        return utils.set_settings('Lifecycle',
                                  self.client,
                                  self.NAMESPACES,
                                  settings,
                                  uris.DCIM_LCService,
                                  "DCIM_LCService",
                                  "DCIM:LCService",
                                  '',
                                  wait_for_idrac=False)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
from unittest import mock
import uuid

import lxml.etree
import lxml.objectify
import requests.exceptions
import requests_mock
import six

from dracclient import exceptions
from dracclient.tests import base
from dracclient.tests import utils as test_utils
import dracclient.wsman


class ClientTestCase(base.BaseTest):

    def setUp(self):
        super(ClientTestCase, self).setUp()
        self.client = dracclient.wsman.Client(**test_utils.FAKE_ENDPOINT)

    @requests_mock.Mocker()
    def test_enumerate(self, mock_requests):
        expected_resp = '<result>yay!</result>'
        mock_requests.post('https://1.2.3.4:443/wsman', text=expected_resp)

        resp = self.client.enumerate('resource', auto_pull=False)
        self.assertEqual('yay!', resp.text)

    def test_enumerate_with_request_failure(self):
        self.client = dracclient.wsman.Client('malformed://^@*', 'user',
                                              'pass')

        self.assertRaises(exceptions.WSManRequestFailure,
                          self.client.enumerate, 'resource')

    @requests_mock.Mocker()
    def test_enumerate_with_invalid_status_code(self, mock_requests):
        mock_requests.post('https://1.2.3.4:443/wsman', status_code=500,
                           reason='dumb request')

        self.assertRaises(exceptions.WSManInvalidResponse,
                          self.client.enumerate, 'resource')

    @requests_mock.Mocker()
    def test_enumerate_with_invalid_utf8(self, mock_requests):
        mock_requests.post('https://1.2.3.4:443/wsman',
                           content=six.b('<result>yay!\xC0</result>'))

        resp = self.client.enumerate('resource')
        self.assertEqual('yay!', resp.text)

    @requests_mock.Mocker()
    def test_enumerate_with_auto_pull(self, mock_requests):
        mock_requests.post(
            'https://1.2.3.4:443/wsman',
            [{'text': test_utils.WSManEnumerations['context'][0]},
             {'text': test_utils.WSManEnumerations['context'][1]},
             {'text': test_utils.WSManEnumerations['context'][2]},
             {'text': test_utils.WSManEnumerations['context'][3]}])

        resp_xml = self.client.enumerate('FooResource')

        foo_resource_uri = 'http://FooResource'
        bar_resource_uri = 'http://BarResource'
        self.assertEqual(
            4, len(resp_xml.findall('.//{%s}FooResource' % foo_resource_uri)))
        self.assertEqual(
            1, len(resp_xml.findall('.//{%s}BazResource' % bar_resource_uri)))
        self.assertEqual(
            0, len(resp_xml.findall(
                './/{%s}EnumerationContext' % dracclient.wsman.NS_WSMAN_ENUM)))

    @requests_mock.Mocker()
    @mock.patch.object(dracclient.wsman.Client, 'pull', autospec=True)
    def test_enumerate_with_auto_pull_without_optimization(self, mock_requests,
                                                           mock_pull):
        mock_requests.post('https://1.2.3.4:443/wsman',
                           text=test_utils.WSManEnumerations['context'][0])
        mock_pull.return_value = lxml.etree.fromstring(
            test_utils.WSManEnumerations['context'][3])

        self.client.enumerate('FooResource', optimization=False, max_elems=42)

        mock_pull.assert_called_once_with(self.client, 'FooResource',
                                          'enum-context-uuid', 42)

    @requests_mock.Mocker()
    def test_pull(self, mock_requests):
        expected_resp = '<result>yay!</result>'
        mock_requests.post('https://1.2.3.4:443/wsman', text=expected_resp)

        resp = self.client.pull('resource', 'context-uuid')

        self.assertEqual('yay!', resp.text)

    @requests_mock.Mocker()
    def test_invoke(self, mock_requests):
        expected_resp = '<result>yay!</result>'
        mock_requests.post('https://1.2.3.4:443/wsman', text=expected_resp)

        resp = self.client.invoke('http://resource', 'method',
                                  {'selector': 'foo'}, {'property': 'bar'})

        self.assertEqual('yay!', resp.text)

    @requests_mock.Mocker()
    def test_invoke_with_ssl_errors(self, mock_requests):
        mock_requests.post('https://1.2.3.4:443/wsman',
                           exc=requests.exceptions.SSLError)

        self.assertRaises(exceptions.WSManRequestFailure,
                          self.client.invoke, 'http://resource', 'method',
                          {'selector': 'foo'}, {'property': 'bar'})

    @requests_mock.Mocker()
    def test_invoke_with_ssl_error_success(self, mock_requests):
        expected_resp = '<result>yay!</result>'
        mock_requests.post('https://1.2.3.4:443/wsman',
                           [{'exc': requests.exceptions.SSLError},
                            {'text': expected_resp}])

        resp = self.client.invoke('http://resource', 'method',
                                  {'selector': 'foo'}, {'property': 'bar'})

        self.assertEqual('yay!', resp.text)

    @requests_mock.Mocker()
    def test_invoke_with_connection_errors(self, mock_requests):
        mock_requests.post('https://1.2.3.4:443/wsman',
                           exc=requests.exceptions.ConnectionError)

        self.assertRaises(exceptions.WSManRequestFailure,
                          self.client.invoke, 'http://resource', 'method',
                          {'selector': 'foo'}, {'property': 'bar'})

    @requests_mock.Mocker()
    def test_invoke_with_connection_error_success(self, mock_requests):
        expected_resp = '<result>yay!</result>'
        mock_requests.post('https://1.2.3.4:443/wsman',
                           [{'exc': requests.exceptions.ConnectionError},
                            {'text': expected_resp}])

        resp = self.client.invoke('http://resource', 'method',
                                  {'selector': 'foo'}, {'property': 'bar'})

        self.assertEqual('yay!', resp.text)

    @requests_mock.Mocker()
    def test_invoke_with_unknown_error(self, mock_requests):
        mock_requests.post('https://1.2.3.4:443/wsman',
                           exc=requests.exceptions.HTTPError)
        self.assertRaises(exceptions.WSManRequestFailure,
                          self.client.invoke, 'http://resource', 'method',
                          {'selector': 'foo'}, {'property': 'bar'})

    @requests_mock.Mocker()
    @mock.patch('time.sleep', autospec=True)
    def test_client_retry_delay(self, mock_requests, mock_ts):
        ssl_retry_delay = 5
        fake_endpoint = test_utils.FAKE_ENDPOINT.copy()
        fake_endpoint['ssl_retry_delay'] = ssl_retry_delay
        client = dracclient.wsman.Client(**fake_endpoint)
        expected_resp = '<result>yay!</result>'
        mock_requests.post('https://1.2.3.4:443/wsman',
                           [{'exc': requests.exceptions.SSLError},
                            {'text': expected_resp}])

        resp = client.invoke('http://resource', 'method',
                             {'selector': 'foo'}, {'property': 'bar'})

        self.assertEqual('yay!', resp.text)
        mock_ts.assert_called_once_with(ssl_retry_delay)


class PayloadTestCase(base.BaseTest):

    def setUp(self):
        super(PayloadTestCase, self).setUp()
        dracclient.wsman.NS_MAP = collections.OrderedDict([
            ('s', dracclient.wsman.NS_SOAP_ENV),
            ('wsa', dracclient.wsman.NS_WS_ADDR),
            ('wsman', dracclient.wsman.NS_WSMAN)])

    @mock.patch.object(uuid, 'uuid4', autospec=True)
    def test_build_enum(self, mock_uuid):
        expected_payload = """<?xml version="1.0" ?>
<s:Envelope xmlns:s="http://www.w3.org/2003/05/soap-envelope"
            xmlns:wsa="http://schemas.xmlsoap.org/ws/2004/08/addressing"
            xmlns:wsman="http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd">
    <s:Header>
        <wsa:To s:mustUnderstand="true">http://host:443/wsman</wsa:To>
        <wsman:ResourceURI s:mustUnderstand="true">http://resource_uri</wsman:ResourceURI>
        <wsa:MessageID s:mustUnderstand="true">uuid:1234-12</wsa:MessageID>
        <wsa:ReplyTo>
            <wsa:Address>http://schemas.xmlsoap.org/ws/2004/08/addressing/role/anonymous</wsa:Address>
        </wsa:ReplyTo>
        <wsa:Action s:mustUnderstand="true">http://schemas.xmlsoap.org/ws/2004/09/enumeration/Enumerate</wsa:Action>
    </s:Header>
    <s:Body>
        <wsen:Enumerate xmlns:wsen="http://schemas.xmlsoap.org/ws/2004/09/enumeration">
            <wsman:OptimizeEnumeration/>
            <wsman:MaxElements>100</wsman:MaxElements>
        </wsen:Enumerate>
    </s:Body>
</s:Envelope>
"""  # noqa
        expected_payload_obj = lxml.objectify.fromstring(expected_payload)

        mock_uuid.return_value = '1234-12'
        payload = dracclient.wsman._EnumeratePayload(
            'http://host:443/wsman', 'http://resource_uri').build()
        payload_obj = lxml.objectify.fromstring(payload)

        self.assertEqual(lxml.etree.tostring(expected_payload_obj),
                         lxml.etree.tostring(payload_obj))

    def test_enumerate_without_optimization(self):
        payload = dracclient.wsman._EnumeratePayload(
            'http://host:443/wsman', 'http://resource_uri', optimization=False,
            max_elems=42).build()
        payload_xml = lxml.etree.fromstring(payload)

        optimize_enum_elems = payload_xml.findall(
            './/{%s}OptimizeEnumeration' % dracclient.wsman.NS_WSMAN)
        max_elem_elems = payload_xml.findall(
            './/{%s}MaxElements' % dracclient.wsman.NS_WSMAN)
        self.assertEqual([], optimize_enum_elems)
        self.assertEqual([], max_elem_elems)

    @mock.patch.object(uuid, 'uuid4', autospec=True)
    def test_build_enum_with_filter(self, mock_uuid):
        expected_payload = """<?xml version="1.0" ?>
<s:Envelope xmlns:s="http://www.w3.org/2003/05/soap-envelope"
            xmlns:wsa="http://schemas.xmlsoap.org/ws/2004/08/addressing"
            xmlns:wsman="http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd">
    <s:Header>
        <wsa:To s:mustUnderstand="true">http://host:443/wsman</wsa:To>
        <wsman:ResourceURI s:mustUnderstand="true">http://resource_uri</wsman:ResourceURI>
        <wsa:MessageID s:mustUnderstand="true">uuid:1234-12</wsa:MessageID>
        <wsa:ReplyTo>
            <wsa:Address>http://schemas.xmlsoap.org/ws/2004/08/addressing/role/anonymous</wsa:Address>
        </wsa:ReplyTo>
        <wsa:Action s:mustUnderstand="true">http://schemas.xmlsoap.org/ws/2004/09/enumeration/Enumerate</wsa:Action>
    </s:Header>
    <s:Body>
        <wsen:Enumerate xmlns:wsen="http://schemas.xmlsoap.org/ws/2004/09/enumeration">
            <wsman:Filter Dialect="http://schemas.dmtf.org/wbem/cql/1/dsp0202.pdf">DROP TABLE users</wsman:Filter>
            <wsman:OptimizeEnumeration/>
            <wsman:MaxElements>100</wsman:MaxElements>
        </wsen:Enumerate>
    </s:Body>
</s:Envelope>
"""  # noqa
        expected_payload_obj = lxml.objectify.fromstring(expected_payload)

        mock_uuid.return_value = '1234-12'
        payload = dracclient.wsman._EnumeratePayload(
            'http://host:443/wsman', 'http://resource_uri',
            filter_query='DROP TABLE users', filter_dialect='cql').build()
        payload_obj = lxml.objectify.fromstring(payload)

        self.assertEqual(lxml.etree.tostring(expected_payload_obj),
                         lxml.etree.tostring(payload_obj))

    def test_build_enum_with_invalid_filter_dialect(self):
        invalid_dialect = 'foo'
        self.assertRaises(exceptions.WSManInvalidFilterDialect,
                          dracclient.wsman._EnumeratePayload,
                          'http://host:443/wsman', 'http://resource_uri',
                          filter_query='DROP TABLE users',
                          filter_dialect=invalid_dialect)

    @mock.patch.object(uuid, 'uuid4', autospec=True)
    def test_build_pull(self, mock_uuid):
        expected_payload = """<?xml version="1.0" ?>
<s:Envelope xmlns:s="http://www.w3.org/2003/05/soap-envelope"
            xmlns:wsa="http://schemas.xmlsoap.org/ws/2004/08/addressing"
            xmlns:wsman="http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd">
    <s:Header>
        <wsa:To s:mustUnderstand="true">http://host:443/wsman</wsa:To>
        <wsman:ResourceURI s:mustUnderstand="true">http://resource_uri</wsman:ResourceURI>
        <wsa:MessageID s:mustUnderstand="true">uuid:1234-12</wsa:MessageID>
        <wsa:ReplyTo>
            <wsa:Address>http://schemas.xmlsoap.org/ws/2004/08/addressing/role/anonymous</wsa:Address>
        </wsa:ReplyTo>
        <wsa:Action s:mustUnderstand="true">http://schemas.xmlsoap.org/ws/2004/09/enumeration/Pull</wsa:Action>
    </s:Header>
    <s:Body>
        <wsen:Pull xmlns:wsen="http://schemas.xmlsoap.org/ws/2004/09/enumeration">
            <wsen:EnumerationContext>context-uuid</wsen:EnumerationContext>
            <wsman:MaxElements>100</wsman:MaxElements>
        </wsen:Pull>
    </s:Body>
</s:Envelope>
"""  # noqa
        expected_payload_obj = lxml.objectify.fromstring(expected_payload)

        mock_uuid.return_value = '1234-12'
        payload = dracclient.wsman._PullPayload('http://host:443/wsman',
                                                'http://resource_uri',
                                                'context-uuid').build()
        payload_obj = lxml.objectify.fromstring(payload)

        self.assertEqual(lxml.etree.tostring(expected_payload_obj),
                         lxml.etree.tostring(payload_obj))

    @mock.patch.object(uuid, 'uuid4', autospec=True)
    def test_build_invoke(self, mock_uuid):
        expected_payload = """<?xml version="1.0" ?>
<s:Envelope xmlns:s="http://www.w3.org/2003/05/soap-envelope"
            xmlns:wsa="http://schemas.xmlsoap.org/ws/2004/08/addressing"
            xmlns:wsman="http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd">
    <s:Header>
        <wsa:To s:mustUnderstand="true">http://host:443/wsman</wsa:To>
        <wsman:ResourceURI s:mustUnderstand="true">http://resource_uri</wsman:ResourceURI>
        <wsa:MessageID s:mustUnderstand="true">uuid:1234-12</wsa:MessageID>
        <wsa:ReplyTo>
            <wsa:Address>http://schemas.xmlsoap.org/ws/2004/08/addressing/role/anonymous</wsa:Address>
        </wsa:ReplyTo>
        <wsa:Action s:mustUnderstand="true">http://resource_uri/method</wsa:Action>
        <wsman:SelectorSet>
            <wsman:Selector Name="selector">foo</wsman:Selector>
        </wsman:SelectorSet>
    </s:Header>
    <s:Body>
        <ns0:method_INPUT xmlns:ns0="http://resource_uri">
            <ns0:property>bar</ns0:property>
        </ns0:method_INPUT>
    </s:Body>
</s:Envelope>
"""  # noqa
        expected_payload_obj = lxml.objectify.fromstring(expected_payload)

        mock_uuid.return_value = '1234-12'
        payload = dracclient.wsman._InvokePayload(
            'http://host:443/wsman', 'http://resource_uri', 'method',
            {'selector': 'foo'}, {'property': 'bar'}).build()
        payload_obj = lxml.objectify.fromstring(payload)

        self.assertEqual(lxml.etree.tostring(expected_payload_obj),
                         lxml.etree.tostring(payload_obj))

    @mock.patch.object(uuid, 'uuid4', autospec=True)
    def test_build_invoke_with_list_in_properties(self, mock_uuid):
        expected_payload = """<?xml version="1.0" ?>
<s:Envelope xmlns:s="http://www.w3.org/2003/05/soap-envelope"
            xmlns:wsa="http://schemas.xmlsoap.org/ws/2004/08/addressing"
            xmlns:wsman="http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd">
    <s:Header>
        <wsa:To s:mustUnderstand="true">http://host:443/wsman</wsa:To>
        <wsman:ResourceURI s:mustUnderstand="true">http://resource_uri</wsman:ResourceURI>
        <wsa:MessageID s:mustUnderstand="true">uuid:1234-12</wsa:MessageID>
        <wsa:ReplyTo>
            <wsa:Address>http://schemas.xmlsoap.org/ws/2004/08/addressing/role/anonymous</wsa:Address>
        </wsa:ReplyTo>
        <wsa:Action s:mustUnderstand="true">http://resource_uri/method</wsa:Action>
        <wsman:SelectorSet>
            <wsman:Selector Name="selector">foo</wsman:Selector>
        </wsman:SelectorSet>
    </s:Header>
    <s:Body>
        <ns0:method_INPUT xmlns:ns0="http://resource_uri">
            <ns0:property>foo</ns0:property>
            <ns0:property>bar</ns0:property>
            <ns0:property>baz</ns0:property>
        </ns0:method_INPUT>
    </s:Body>
</s:Envelope>
"""  # noqa
        expected_payload_obj = lxml.objectify.fromstring(expected_payload)

        mock_uuid.return_value = '1234-12'
        payload = dracclient.wsman._InvokePayload(
            'http://host:443/wsman', 'http://resource_uri', 'method',
            {'selector': 'foo'}, {'property': ['foo', 'bar', 'baz']}).build()
        payload_obj = lxml.objectify.fromstring(payload)

        self.assertEqual(lxml.etree.tostring(expected_payload_obj),
                         lxml.etree.tostring(payload_obj))



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# iDRAC is ready retry constants
DEFAULT_IDRAC_IS_READY_RETRIES = 96
DEFAULT_IDRAC_IS_READY_RETRY_DELAY_SEC = 10

# Web Services Management (WS-Management and WS-Man) SSL retry on error
# behavior constants
DEFAULT_WSMAN_SSL_ERROR_RETRIES = 3
DEFAULT_WSMAN_SSL_ERROR_RETRY_DELAY_SEC = 0

NOT_SUPPORTED_MSG = " operation is not supported on th"

# power states
POWER_ON = 'POWER_ON'
POWER_OFF = 'POWER_OFF'
REBOOT = 'REBOOT'

PRIMARY_STATUS = {
    '0': 'unknown',
    '1': 'ok',
    '2': 'degraded',
    '3': 'error'
}

# binary unit constants
UNITS_KI = 2 ** 10

# Lifecycle Controller status constant
LC_IN_RECOVERY = '4'


# Reboot required indicator
# Note: When the iDRAC returns optional for this value, this indicates that
# either a reboot may be performed to complete the operation or a real time
# config job may be created to complete the operation without performing a
# reboot.  This library does not currently support creating a real time config
# job, so a reboot must be performed when a value of "optional" is returned.
class RebootRequired(object):
    true = 'true'
    optional = 'optional'
    false = 'false'

    @classmethod
    def all(cls):
        return [cls.true, cls.optional, cls.false]


class RebootJobType(object):
    """Enumeration of different reboot job types."""

    power_cycle = 'power_cycle'
    """Hard power off, power on cycle"""

    graceful_reboot = 'graceful_reboot'
    """OS level reboot and wait for OS shutdown"""

    reboot_forced_shutdown = 'reboot_forced_shutdown'
    """OS level reboot and force power cycle if OS does not shut
    down
    """

    @classmethod
    def all(cls):
        return [cls.power_cycle,
                cls.graceful_reboot,
                cls.reboot_forced_shutdown]


class RaidStatus(object):
    """Enumeration of different volume types."""

    jbod = 'JBOD'
    """Just a Bunch of Disks"""

    raid = 'RAID'
    """Redundant Array of Independent Disks"""

    @classmethod
    def all(cls):
        return [cls.jbod,
                cls.raid]



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import requests_mock

import dracclient.client
from dracclient import constants
from dracclient import exceptions
from dracclient.resources import uris
from dracclient.tests import base
from dracclient.tests import utils as test_utils


@requests_mock.Mocker()
class WSManClientTestCase(base.BaseTest):

    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    def test_enumerate(self, mock_requests, mock_wait_until_idrac_is_ready):
        mock_requests.post('https://1.2.3.4:443/wsman',
                           text='<result>yay!</result>')

        client = dracclient.client.WSManClient(**test_utils.FAKE_ENDPOINT)
        resp = client.enumerate('http://resource')
        mock_wait_until_idrac_is_ready.assert_called_once_with(client)
        self.assertEqual('yay!', resp.text)

    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    def test_enumerate_without_wait_for_idrac(
            self, mock_requests, mock_wait_until_idrac_is_ready):
        mock_requests.post('https://1.2.3.4:443/wsman',
                           text='<result>yay!</result>')

        client = dracclient.client.WSManClient(**test_utils.FAKE_ENDPOINT)
        resp = client.enumerate('http://resource', wait_for_idrac=False)
        self.assertFalse(mock_wait_until_idrac_is_ready.called)
        self.assertEqual('yay!', resp.text)

    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    def test_invoke(self, mock_requests, mock_wait_until_idrac_is_ready):
        xml = """
<response xmlns:n1="http://resource">
    <n1:ReturnValue>42</n1:ReturnValue>
    <result>yay!</result>
</response>
"""  # noqa
        mock_requests.post('https://1.2.3.4:443/wsman', text=xml)

        client = dracclient.client.WSManClient(**test_utils.FAKE_ENDPOINT)
        resp = client.invoke('http://resource', 'Foo')
        mock_wait_until_idrac_is_ready.assert_called_once_with(client)
        self.assertEqual('yay!', resp.find('result').text)

    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    def test_invoke_without_wait_for_idrac(
            self, mock_requests, mock_wait_until_idrac_is_ready):
        xml = """
<response xmlns:n1="http://resource">
    <n1:ReturnValue>42</n1:ReturnValue>
    <result>yay!</result>
</response>
"""  # noqa
        mock_requests.post('https://1.2.3.4:443/wsman', text=xml)

        client = dracclient.client.WSManClient(**test_utils.FAKE_ENDPOINT)
        resp = client.invoke('http://resource', 'Foo', wait_for_idrac=False)
        self.assertFalse(mock_wait_until_idrac_is_ready.called)
        self.assertEqual('yay!', resp.find('result').text)

    def test_invoke_with_expected_return_value(self, mock_requests):
        xml = """
<response xmlns:n1="http://resource">
    <n1:ReturnValue>42</n1:ReturnValue>
    <result>yay!</result>
</response>
"""  # noqa
        mock_requests.post('https://1.2.3.4:443/wsman', text=xml)

        client = dracclient.client.WSManClient(**test_utils.FAKE_ENDPOINT)
        resp = client.invoke('http://resource', 'Foo',
                             expected_return_value='42', wait_for_idrac=False)
        self.assertEqual('yay!', resp.find('result').text)

    def test_invoke_with_error_return_value(self, mock_requests):
        xml = """
<response xmlns:n1="http://resource">
    <n1:ReturnValue>2</n1:ReturnValue>
    <result>yay!</result>
</response>
"""  # noqa
        mock_requests.post('https://1.2.3.4:443/wsman', text=xml)

        client = dracclient.client.WSManClient(**test_utils.FAKE_ENDPOINT)
        self.assertRaises(exceptions.DRACOperationFailed, client.invoke,
                          'http://resource', 'Foo', wait_for_idrac=False)

    def test_invoke_with_unchecked_return_value(self, mock_requests):
        xml = """
<response xmlns:n1="http://resource">
    <n1:ReturnValue>2</n1:ReturnValue>
    <result>yay!</result>
</response>
"""  # noqa
        mock_requests.post('https://1.2.3.4:443/wsman', text=xml)

        client = dracclient.client.WSManClient(**test_utils.FAKE_ENDPOINT)
        resp = client.invoke('http://resource', 'Foo',
                             wait_for_idrac=False, check_return_value=False)
        self.assertEqual('yay!', resp.find('result').text)

    def test_invoke_with_unexpected_return_value(self, mock_requests):
        xml = """
<response xmlns:n1="http://resource">
    <n1:ReturnValue>42</n1:ReturnValue>
    <result>yay!</result>
</response>
"""  # noqa
        mock_requests.post('https://1.2.3.4:443/wsman', text=xml)

        client = dracclient.client.WSManClient(**test_utils.FAKE_ENDPOINT)
        self.assertRaises(exceptions.DRACUnexpectedReturnValue, client.invoke,
                          'http://resource', 'Foo',
                          expected_return_value='4242', wait_for_idrac=False)

    def test_is_idrac_ready_ready(self, mock_requests):
        expected_text = test_utils.LifecycleControllerInvocations[
            uris.DCIM_LCService]['GetRemoteServicesAPIStatus']['is_ready']
        mock_requests.post('https://1.2.3.4:443/wsman',
                           text=expected_text)

        client = dracclient.client.WSManClient(**test_utils.FAKE_ENDPOINT)
        self.assertTrue(client.is_idrac_ready())

    def test_is_idrac_ready_not_ready(self, mock_requests):
        expected_text = test_utils.LifecycleControllerInvocations[
            uris.DCIM_LCService]['GetRemoteServicesAPIStatus']['is_not_ready']
        mock_requests.post('https://1.2.3.4:443/wsman',
                           text=expected_text)

        client = dracclient.client.WSManClient(**test_utils.FAKE_ENDPOINT)
        self.assertFalse(client.is_idrac_ready())

    @mock.patch.object(dracclient.client.WSManClient, 'is_idrac_ready',
                       autospec=True)
    @mock.patch('time.sleep', autospec=True)
    def test_wait_until_idrac_is_ready_with_none_arguments(
            self, mock_requests, mock_ts, mock_is_idrac_ready):
        ready_retries = 2
        ready_retry_delay = 1

        side_effect = (ready_retries - 1) * [False]
        side_effect.append(True)
        mock_is_idrac_ready.side_effect = side_effect

        fake_endpoint = test_utils.FAKE_ENDPOINT.copy()
        fake_endpoint['ready_retries'] = ready_retries
        fake_endpoint['ready_retry_delay'] = ready_retry_delay

        client = dracclient.client.WSManClient(**fake_endpoint)
        client.wait_until_idrac_is_ready(retries=None, retry_delay=None)

        self.assertEqual(mock_is_idrac_ready.call_count, ready_retries)
        self.assertEqual(mock_ts.call_count, ready_retries - 1)
        mock_ts.assert_called_with(ready_retry_delay)

    @mock.patch.object(dracclient.client.WSManClient, 'is_idrac_ready',
                       autospec=True)
    @mock.patch('time.sleep', autospec=True)
    def test_wait_until_idrac_is_ready_with_non_none_arguments(
            self, mock_requests, mock_ts, mock_is_idrac_ready):
        retries = 2
        self.assertNotEqual(retries, constants.DEFAULT_IDRAC_IS_READY_RETRIES)

        retry_delay = 1
        self.assertNotEqual(
            retry_delay, constants.DEFAULT_IDRAC_IS_READY_RETRY_DELAY_SEC)

        side_effect = (retries - 1) * [False]
        side_effect.append(True)
        mock_is_idrac_ready.side_effect = side_effect

        fake_endpoint = test_utils.FAKE_ENDPOINT.copy()
        fake_endpoint['ready_retries'] = (
            constants.DEFAULT_IDRAC_IS_READY_RETRIES)
        fake_endpoint['ready_retry_delay'] = (
            constants.DEFAULT_IDRAC_IS_READY_RETRY_DELAY_SEC)

        client = dracclient.client.WSManClient(**fake_endpoint)
        client.wait_until_idrac_is_ready(retries, retry_delay)

        self.assertEqual(mock_is_idrac_ready.call_count, retries)
        self.assertEqual(mock_ts.call_count, retries - 1)
        mock_ts.assert_called_with(retry_delay)

    def test_wait_until_idrac_is_ready_ready(self, mock_requests):
        expected_text = test_utils.LifecycleControllerInvocations[
            uris.DCIM_LCService]['GetRemoteServicesAPIStatus']['is_ready']
        mock_requests.post('https://1.2.3.4:443/wsman',
                           text=expected_text)

        client = dracclient.client.WSManClient(**test_utils.FAKE_ENDPOINT)

        try:
            client.wait_until_idrac_is_ready()
        except exceptions.DRACOperationFailed:
            self.fail('wait_until_idrac_is_ready() timed out when it should '
                      'not have!')

    @mock.patch('time.sleep', autospec=True)
    def test_wait_until_idrac_is_ready_timeout(self,
                                               mock_requests,
                                               mock_ts):
        expected_text = test_utils.LifecycleControllerInvocations[
            uris.DCIM_LCService]['GetRemoteServicesAPIStatus']['is_not_ready']
        mock_requests.post('https://1.2.3.4:443/wsman',
                           text=expected_text)

        client = dracclient.client.WSManClient(**test_utils.FAKE_ENDPOINT)
        self.assertRaises(exceptions.DRACOperationFailed,
                          client.wait_until_idrac_is_ready)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections

from dracclient import constants
from dracclient.resources import uris
from dracclient import utils

CPU_CHARACTERISTICS_64BIT = '4'

NIC_LINK_SPEED_MBPS = {
    '0': None,
    '1': 10,
    '2': 100,
    '3': 1000,
    '4': 2.5 * constants.UNITS_KI,
    '5': 10 * constants.UNITS_KI,
    '6': 20 * constants.UNITS_KI,
    '7': 40 * constants.UNITS_KI,
    '8': 100 * constants.UNITS_KI,
    '9': 25 * constants.UNITS_KI,
    '10': 50 * constants.UNITS_KI
}

NIC_LINK_DUPLEX = {
    '0': 'unknown',
    '1': 'full duplex',
    '2': 'half duplex'
}

NIC_MODE = {
    '0': 'unknown',
    '2': 'enabled',
    '3': 'disabled'}

CPU = collections.namedtuple(
    'CPU',
    ['id', 'cores', 'speed_mhz', 'model', 'status', 'ht_enabled',
     'cpu_count', 'turbo_enabled', 'vt_enabled', 'arch64'])

Memory = collections.namedtuple(
    'Memory',
    ['id', 'size_mb', 'speed_mhz', 'manufacturer', 'model', 'status'])

NIC = collections.namedtuple(
    'NIC',
    ['id', 'mac', 'model', 'speed_mbps', 'duplex', 'media_type'])

Video = collections.namedtuple(
    'Video',
    ['id', 'description', 'function_number', 'manufacturer', 'pci_device_id',
     'pci_vendor_id', 'pci_subdevice_id', 'pci_subvendor_id'])

System = collections.namedtuple(
    'System',
    ['id', 'lcc_version', 'model', 'service_tag', 'uuid',
     'last_system_inventory_time'])


class InventoryManagement(object):

    def __init__(self, client):
        """Creates InventoryManagement object

        :param client: an instance of WSManClient
        """
        self.client = client

    def list_cpus(self):
        """Returns the list of CPUs

        :returns: a list of CPU objects
        :raises: WSManRequestFailure on request failures
        :raises: WSManInvalidResponse when receiving invalid response
        :raises: DRACOperationFailed on error reported back by the DRAC
        """

        doc = self.client.enumerate(uris.DCIM_CPUView)

        cpus = utils.find_xml(doc, 'DCIM_CPUView',
                              uris.DCIM_CPUView,
                              find_all=True)

        return [self._parse_cpus(cpu) for cpu in cpus]

    def _parse_cpus(self, cpu):
        drac_characteristics = self._get_cpu_attr(cpu, 'Characteristics')
        arch64 = (CPU_CHARACTERISTICS_64BIT == drac_characteristics)

        return CPU(
            id=self._get_cpu_attr(cpu, 'FQDD'),
            cores=int(self._get_cpu_attr(cpu, 'NumberOfProcessorCores')),
            speed_mhz=int(self._get_cpu_attr(cpu, 'CurrentClockSpeed')),
            model=self._get_cpu_attr(cpu, 'Model'),
            status=constants.PRIMARY_STATUS[
                self._get_cpu_attr(cpu, 'PrimaryStatus')],
            ht_enabled=bool(self._get_cpu_attr(cpu, 'HyperThreadingEnabled',
                                               allow_missing=True)),
            turbo_enabled=bool(self._get_cpu_attr(cpu, 'TurboModeEnabled',
                                                  allow_missing=True)),
            cpu_count=self._get_cpu_count(
                int(self._get_cpu_attr(cpu, 'NumberOfProcessorCores')),
                bool(self._get_cpu_attr(cpu, 'HyperThreadingEnabled',
                                        allow_missing=True))),
            vt_enabled=bool(self._get_cpu_attr(
                cpu, 'VirtualizationTechnologyEnabled', allow_missing=True)),
            arch64=arch64)

    def _get_cpu_count(self, cores, ht_enabled):
        if ht_enabled:
            return int(cores * 2)
        else:
            return int(cores)

    def _get_cpu_attr(self, cpu, attr_name, allow_missing=False):
        return utils.get_wsman_resource_attr(
            cpu, uris.DCIM_CPUView, attr_name, allow_missing=allow_missing)

    def list_memory(self):
        """Returns the list of installed memory

        :returns: a list of Memory objects
        :raises: WSManRequestFailure on request failures
        :raises: WSManInvalidResponse when receiving invalid response
        :raises: DRACOperationFailed on error reported back by the DRAC
        """

        doc = self.client.enumerate(uris.DCIM_MemoryView)

        installed_memory = utils.find_xml(doc, 'DCIM_MemoryView',
                                          uris.DCIM_MemoryView,
                                          find_all=True)

        return [self._parse_memory(memory) for memory in installed_memory]

    def _parse_memory(self, memory):
        return Memory(
            id=self._get_memory_attr(memory, 'FQDD'),
            size_mb=int(self._get_memory_attr(memory, 'Size')),
            speed_mhz=int(self._get_memory_attr(memory, 'Speed')),
            manufacturer=self._get_memory_attr(memory, 'Manufacturer'),
            model=self._get_memory_attr(memory, 'Model'),
            status=constants.PRIMARY_STATUS[
                self._get_memory_attr(memory, 'PrimaryStatus')])

    def _get_memory_attr(self, memory, attr_name):
        return utils.get_wsman_resource_attr(memory, uris.DCIM_MemoryView,
                                             attr_name)

    def list_nics(self, sort=False):
        """Returns the list of NICs

        :returns: a list of NIC objects
        :raises: WSManRequestFailure on request failures
        :raises: WSManInvalidResponse when receiving invalid response
        :raises: DRACOperationFailed on error reported back by the DRAC
                 interface
        """

        doc = self.client.enumerate(uris.DCIM_NICView)
        drac_nics = utils.find_xml(doc, 'DCIM_NICView', uris.DCIM_NICView,
                                   find_all=True)
        nics = [self._parse_drac_nic(nic) for nic in drac_nics]
        if sort:
            nics.sort(key=lambda nic: nic.id)

        return nics

    def _parse_drac_nic(self, drac_nic):
        fqdd = self._get_nic_attr(drac_nic, 'FQDD')
        drac_speed = self._get_nic_attr(drac_nic, 'LinkSpeed')
        drac_duplex = self._get_nic_attr(drac_nic, 'LinkDuplex')

        return NIC(
            id=fqdd,
            mac=self._get_nic_attr(drac_nic, 'CurrentMACAddress'),
            model=self._get_nic_attr(drac_nic, 'ProductName'),
            speed_mbps=NIC_LINK_SPEED_MBPS[drac_speed],
            duplex=NIC_LINK_DUPLEX[drac_duplex],
            media_type=self._get_nic_attr(drac_nic, 'MediaType'))

    def _get_nic_attr(self, drac_nic, attr_name):
        return utils.get_wsman_resource_attr(drac_nic, uris.DCIM_NICView,
                                             attr_name)

    def list_video_controllers(self):
        """Returns the list of video controllers

        :returns: a list of Video objects
        :raises: WSManRequestFailure on request failures
        :raises: WSManInvalidResponse when receiving invalid response
        :raises: DRACOperationFailed on error reported back by the DRAC
        """

        doc = self.client.enumerate(uris.DCIM_VideoView)

        controllers = utils.find_xml(doc, 'DCIM_VideoView',
                                     uris.DCIM_VideoView,
                                     find_all=True)

        return [self._parse_video_controllers(controller)
                for controller in controllers]

    def _parse_video_controllers(self, controller):
        return Video(
            id=self._get_video_attr(controller, 'FQDD'),
            description=self._get_video_attr(controller, 'Description'),
            function_number=int(self._get_video_attr(controller,
                                                     'FunctionNumber')),
            manufacturer=self._get_video_attr(controller, 'Manufacturer'),
            pci_device_id=self._get_video_attr(controller, 'PCIDeviceID'),
            pci_vendor_id=self._get_video_attr(controller, 'PCIVendorID'),
            pci_subdevice_id=self._get_video_attr(controller,
                                                  'PCISubDeviceID'),
            pci_subvendor_id=self._get_video_attr(controller,
                                                  'PCISubVendorID'))

    def _get_video_attr(self, controller, attr_name, allow_missing=False):
        return utils.get_wsman_resource_attr(
            controller, uris.DCIM_VideoView, attr_name,
            allow_missing=allow_missing)

    def get_system(self):
        """Returns a System object

            :returns: a System object
            :raises: WSManRequestFailure on request failures
            :raises: WSManInvalidRespons when receiving invalid response
        """
        doc = self.client.enumerate(uris.DCIM_SystemView)
        drac_system = utils.find_xml(doc,
                                     'DCIM_SystemView',
                                     uris.DCIM_SystemView,
                                     find_all=False)

        return self._parse_drac_system(drac_system)

    def _parse_drac_system(self, drac_system):
        return System(
            id=self._get_system_attr(drac_system, 'InstanceID'),
            uuid=self._get_system_attr(drac_system, 'UUID'),
            service_tag=self._get_system_attr(drac_system, 'ServiceTag'),
            model=self._get_system_attr(drac_system, 'Model'),
            lcc_version=self._get_system_attr(drac_system,
                                              'LifecycleControllerVersion'),
            last_system_inventory_time=self._get_system_attr(
                drac_system, 'LastSystemInventoryTime').split('.')[0])

    def _get_system_attr(self, drac_system, attr_name):
        return utils.get_wsman_resource_attr(drac_system,
                                             uris.DCIM_SystemView,
                                             attr_name)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import lxml.etree
from unittest import mock

import requests_mock

import dracclient.client
from dracclient import exceptions
import dracclient.resources.job
from dracclient.resources import uris
from dracclient.tests import base
from dracclient.tests import utils as test_utils
from dracclient import utils


class ClientJobManagementTestCase(base.BaseTest):

    def setUp(self):
        super(ClientJobManagementTestCase, self).setUp()
        self.drac_client = dracclient.client.DRACClient(
            **test_utils.FAKE_ENDPOINT)

    @requests_mock.Mocker()
    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    def test_list_jobs(self, mock_requests, mock_wait_until_idrac_is_ready):
        mock_requests.post(
            'https://1.2.3.4:443/wsman',
            text=test_utils.JobEnumerations[uris.DCIM_LifecycleJob]['ok'])

        jobs = self.drac_client.list_jobs()

        self.assertEqual(6, len(jobs))

    @mock.patch.object(dracclient.client.WSManClient, 'enumerate',
                       spec_set=True, autospec=True)
    def test_list_jobs_only_unfinished(self, mock_enumerate):
        expected_filter_query = ('select * from DCIM_LifecycleJob '
                                 'where Name != "CLEARALL" and '
                                 'JobStatus != "Reboot Completed" and '
                                 'JobStatus != "Reboot Failed" and '
                                 'JobStatus != "Completed" and '
                                 'JobStatus != "Completed with Errors" and '
                                 'JobStatus != "Failed"')
        mock_enumerate.return_value = lxml.etree.fromstring(
            test_utils.JobEnumerations[uris.DCIM_LifecycleJob]['ok'])

        self.drac_client.list_jobs(only_unfinished=True)

        mock_enumerate.assert_called_once_with(
            mock.ANY, uris.DCIM_LifecycleJob,
            filter_query=expected_filter_query)

    @mock.patch.object(dracclient.client.WSManClient, 'enumerate',
                       spec_set=True, autospec=True)
    def test_get_job(self, mock_enumerate):
        expected_filter_query = ('select * from DCIM_LifecycleJob'
                                 ' where InstanceID="42"')
        # NOTE: This is the first job in the xml. Filtering the job is the
        #       responsibility of the controller, so not testing it.
        expected_job = dracclient.resources.job.Job(id='JID_CLEARALL',
                                                    name='CLEARALL',
                                                    start_time='TIME_NA',
                                                    until_time='TIME_NA',
                                                    message='NA',
                                                    status='Pending',
                                                    percent_complete='0')
        mock_enumerate.return_value = lxml.etree.fromstring(
            test_utils.JobEnumerations[uris.DCIM_LifecycleJob]['ok'])

        job = self.drac_client.get_job(42)

        mock_enumerate.assert_called_once_with(
            mock.ANY, uris.DCIM_LifecycleJob,
            filter_query=expected_filter_query)
        self.assertEqual(expected_job, job)

    @mock.patch.object(dracclient.client.WSManClient, 'enumerate',
                       spec_set=True, autospec=True)
    def test_get_job_not_found(self, mock_enumerate):
        expected_filter_query = ('select * from DCIM_LifecycleJob'
                                 ' where InstanceID="42"')
        mock_enumerate.return_value = lxml.etree.fromstring(
            test_utils.JobEnumerations[uris.DCIM_LifecycleJob]['not_found'])

        job = self.drac_client.get_job(42)

        mock_enumerate.assert_called_once_with(
            mock.ANY, uris.DCIM_LifecycleJob,
            filter_query=expected_filter_query)
        self.assertIsNone(job)

    @mock.patch.object(dracclient.client.WSManClient, 'invoke',
                       spec_set=True, autospec=True)
    def test_delete_jobs_all(self, mock_invoke):
        expected_selectors = {'SystemCreationClassName': 'DCIM_ComputerSystem',
                              'SystemName': 'idrac',
                              'CreationClassName': 'DCIM_JobService',
                              'Name': 'JobService'}
        expected_properties = {'JobID': 'JID_CLEARALL'}

        self.drac_client.delete_jobs()

        mock_invoke.assert_called_once_with(
            mock.ANY, uris.DCIM_JobService, 'DeleteJobQueue',
            expected_selectors, expected_properties,
            expected_return_value=utils.RET_SUCCESS)

    @mock.patch.object(dracclient.client.WSManClient, 'invoke',
                       spec_set=True, autospec=True)
    def test_delete_jobs_force(self, mock_invoke):
        expected_selectors = {'SystemCreationClassName': 'DCIM_ComputerSystem',
                              'SystemName': 'idrac',
                              'CreationClassName': 'DCIM_JobService',
                              'Name': 'JobService'}
        expected_properties = {'JobID': 'JID_CLEARALL_FORCE'}

        self.drac_client.delete_jobs(['JID_CLEARALL_FORCE'])

        mock_invoke.assert_called_once_with(
            mock.ANY, uris.DCIM_JobService, 'DeleteJobQueue',
            expected_selectors, expected_properties,
            expected_return_value=utils.RET_SUCCESS)

    @mock.patch.object(dracclient.client.WSManClient, 'invoke',
                       spec_set=True, autospec=True)
    def test_delete_jobs_one(self, mock_invoke):
        expected_selectors = {'SystemCreationClassName': 'DCIM_ComputerSystem',
                              'SystemName': 'idrac',
                              'CreationClassName': 'DCIM_JobService',
                              'Name': 'JobService'}
        expected_properties = {'JobID': 'JID_442507917525'}

        self.drac_client.delete_jobs(['JID_442507917525'])

        mock_invoke.assert_called_once_with(
            mock.ANY, uris.DCIM_JobService, 'DeleteJobQueue',
            expected_selectors, expected_properties,
            expected_return_value=utils.RET_SUCCESS)

    @mock.patch.object(dracclient.client.WSManClient, 'invoke',
                       spec_set=True, autospec=True)
    def test_delete_jobs_multi(self, mock_invoke):
        expected_selectors = {'SystemCreationClassName': 'DCIM_ComputerSystem',
                              'SystemName': 'idrac',
                              'CreationClassName': 'DCIM_JobService',
                              'Name': 'JobService'}

        self.drac_client.delete_jobs(['JID_442507917525',
                                      'JID_442507917526'])

        calls_expected = [
            mock.call(mock.ANY,
                      uris.DCIM_JobService,
                      'DeleteJobQueue',
                      expected_selectors,
                      {'JobID': 'JID_442507917525'},
                      expected_return_value=utils.RET_SUCCESS),
            mock.call(mock.ANY,
                      uris.DCIM_JobService,
                      'DeleteJobQueue',
                      expected_selectors,
                      {'JobID': 'JID_442507917526'},
                      expected_return_value=utils.RET_SUCCESS)]
        mock_invoke.assert_has_calls(calls_expected)

    @mock.patch.object(dracclient.client.WSManClient, 'invoke',
                       spec_set=True, autospec=True)
    def test_delete_jobs_none(self, mock_invoke):
        self.drac_client.delete_jobs(None)
        self.assertFalse(mock_invoke.called)

    @mock.patch.object(dracclient.client.WSManClient, 'invoke',
                       spec_set=True, autospec=True)
    def test_delete_jobs_empty_list(self, mock_invoke):
        self.drac_client.delete_jobs([])
        self.assertFalse(mock_invoke.called)

    @requests_mock.Mocker()
    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    def test_delete_job_not_found(
            self, mock_requests,
            mock_wait_until_idrac_is_ready):
        mock_requests.post(
            'https://1.2.3.4:443/wsman',
            text=test_utils.JobService[uris.DCIM_JobService][
                'DeleteJobQueue']['error'])
        self.assertRaises(
            exceptions.DRACOperationFailed,
            self.drac_client.delete_jobs,
            ['JID_1234'])

    @requests_mock.Mocker()
    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    def test_delete_some_jobs_not_found(
            self, mock_requests,
            mock_wait_until_idrac_is_ready):
        mock_requests.post(
            'https://1.2.3.4:443/wsman',
            [{'text': test_utils.JobService[uris.DCIM_JobService][
                'DeleteJobQueue']['error']},
             {'text': test_utils.JobService[uris.DCIM_JobService][
                 'DeleteJobQueue']['ok']}])

        self.assertRaises(
            exceptions.DRACOperationFailed,
            self.drac_client.delete_jobs,
            ['JID_1234', 'JID_442507917525'])

        self.assertEqual(mock_requests.call_count, 2)

    @mock.patch.object(dracclient.client.WSManClient, 'invoke',
                       spec_set=True, autospec=True)
    def test_create_config_job_for_lifecycle(self, mock_invoke):
        cim_creation_class_name = 'DCIM_LCService'
        cim_name = 'DCIM:LCService'
        target = ''

        expected_selectors = {'CreationClassName': cim_creation_class_name,
                              'Name': cim_name,
                              'SystemCreationClassName': 'DCIM_ComputerSystem',
                              'SystemName': 'DCIM:ComputerSystem'}
        expected_properties = {'Target': target,
                               'ScheduledStartTime': 'TIME_NOW'}

        mock_invoke.return_value = lxml.etree.fromstring(
            test_utils.JobInvocations[uris.DCIM_LCService][
                'CreateConfigJob']['ok'])

        job_id = self.drac_client.create_config_job(
            uris.DCIM_LCService, cim_creation_class_name, cim_name, target,
            start_time='TIME_NOW',
            wait_for_idrac=False, method_name='CreateConfigJob')

        mock_invoke.assert_called_once_with(
            mock.ANY, uris.DCIM_LCService, 'CreateConfigJob',
            expected_selectors, expected_properties,
            expected_return_value=utils.RET_CREATED,
            wait_for_idrac=False)
        self.assertEqual('JID_442507917525', job_id)

    @mock.patch.object(dracclient.client.WSManClient, 'invoke',
                       spec_set=True, autospec=True)
    def test_create_config_job(self, mock_invoke):
        cim_creation_class_name = 'DCIM_BIOSService'
        cim_name = 'DCIM:BIOSService'
        target = 'BIOS.Setup.1-1'
        wait_for_idrac = True
        expected_selectors = {'CreationClassName': cim_creation_class_name,
                              'Name': cim_name,
                              'SystemCreationClassName': 'DCIM_ComputerSystem',
                              'SystemName': 'DCIM:ComputerSystem'}
        expected_properties = {'Target': target,
                               'ScheduledStartTime': 'TIME_NOW'}

        mock_invoke.return_value = lxml.etree.fromstring(
            test_utils.JobInvocations[uris.DCIM_BIOSService][
                'CreateTargetedConfigJob']['ok'])

        job_id = self.drac_client.create_config_job(
            uris.DCIM_BIOSService, cim_creation_class_name, cim_name, target)

        mock_invoke.assert_called_once_with(
            mock.ANY, uris.DCIM_BIOSService, 'CreateTargetedConfigJob',
            expected_selectors, expected_properties,
            expected_return_value=utils.RET_CREATED,
            wait_for_idrac=wait_for_idrac)
        self.assertEqual('JID_442507917525', job_id)

    @mock.patch.object(dracclient.client.WSManClient, 'invoke',
                       spec_set=True, autospec=True)
    def test_create_config_job_with_start_time(self, mock_invoke):
        cim_creation_class_name = 'DCIM_BIOSService'
        cim_name = 'DCIM:BIOSService'
        target = 'BIOS.Setup.1-1'
        start_time = "20140924120105"
        wait_for_idrac = True
        expected_selectors = {'CreationClassName': cim_creation_class_name,
                              'Name': cim_name,
                              'SystemCreationClassName': 'DCIM_ComputerSystem',
                              'SystemName': 'DCIM:ComputerSystem'}
        expected_properties = {'Target': target,
                               'ScheduledStartTime': start_time}
        mock_invoke.return_value = lxml.etree.fromstring(
            test_utils.JobInvocations[uris.DCIM_BIOSService][
                'CreateTargetedConfigJob']['ok'])

        job_id = self.drac_client.create_config_job(
            uris.DCIM_BIOSService, cim_creation_class_name, cim_name, target,
            start_time=start_time)

        mock_invoke.assert_called_once_with(
            mock.ANY, uris.DCIM_BIOSService, 'CreateTargetedConfigJob',
            expected_selectors, expected_properties,
            expected_return_value=utils.RET_CREATED,
            wait_for_idrac=wait_for_idrac)
        self.assertEqual('JID_442507917525', job_id)

    @mock.patch.object(dracclient.client.WSManClient, 'invoke',
                       spec_set=True, autospec=True)
    def test_create_config_job_with_no_start_time(self, mock_invoke):
        cim_creation_class_name = 'DCIM_BIOSService'
        cim_name = 'DCIM:BIOSService'
        target = 'BIOS.Setup.1-1'
        start_time = None
        wait_for_idrac = True
        expected_selectors = {'CreationClassName': cim_creation_class_name,
                              'Name': cim_name,
                              'SystemCreationClassName': 'DCIM_ComputerSystem',
                              'SystemName': 'DCIM:ComputerSystem'}
        expected_properties = {'Target': target}
        mock_invoke.return_value = lxml.etree.fromstring(
            test_utils.JobInvocations[uris.DCIM_BIOSService][
                'CreateTargetedConfigJob']['ok'])

        job_id = self.drac_client.create_config_job(
            uris.DCIM_BIOSService, cim_creation_class_name, cim_name, target,
            start_time=start_time)

        mock_invoke.assert_called_once_with(
            mock.ANY, uris.DCIM_BIOSService, 'CreateTargetedConfigJob',
            expected_selectors, expected_properties,
            expected_return_value=utils.RET_CREATED,
            wait_for_idrac=wait_for_idrac)
        self.assertEqual('JID_442507917525', job_id)

    @requests_mock.Mocker()
    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    def test_create_config_job_failed(self, mock_requests,
                                      mock_wait_until_idrac_is_ready):
        cim_creation_class_name = 'DCIM_BIOSService'
        cim_name = 'DCIM:BIOSService'
        target = 'BIOS.Setup.1-1'
        mock_requests.post(
            'https://1.2.3.4:443/wsman',
            text=test_utils.JobInvocations[uris.DCIM_BIOSService][
                'CreateTargetedConfigJob']['error'])

        self.assertRaises(
            exceptions.DRACOperationFailed, self.drac_client.create_config_job,
            uris.DCIM_BIOSService, cim_creation_class_name, cim_name, target)

    @requests_mock.Mocker()
    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    def test_create_config_job_for_lifecycle_failed(
            self, mock_requests,
            mock_wait_until_idrac_is_ready):
        cim_creation_class_name = 'DCIM_LCService'
        cim_name = 'DCIM:LCService'
        target = ''
        mock_requests.post(
            'https://1.2.3.4:443/wsman',
            text=test_utils.JobInvocations[uris.DCIM_LCService][
                'CreateConfigJob']['error'])

        self.assertRaises(
            exceptions.DRACOperationFailed, self.drac_client.create_config_job,
            uris.DCIM_LCService, cim_creation_class_name, cim_name, target)

    @mock.patch.object(dracclient.client.WSManClient, 'invoke', spec_set=True,
                       autospec=True)
    def test_create_config_job_with_reboot(self, mock_invoke):
        cim_creation_class_name = 'DCIM_BIOSService'
        cim_name = 'DCIM:BIOSService'
        target = 'BIOS.Setup.1-1'
        wait_for_idrac = True
        expected_selectors = {'CreationClassName': cim_creation_class_name,
                              'Name': cim_name,
                              'SystemCreationClassName': 'DCIM_ComputerSystem',
                              'SystemName': 'DCIM:ComputerSystem'}
        expected_properties = {'Target': target,
                               'RebootJobType': '3',
                               'ScheduledStartTime': 'TIME_NOW'}
        mock_invoke.return_value = lxml.etree.fromstring(
            test_utils.JobInvocations[uris.DCIM_BIOSService][
                'CreateTargetedConfigJob']['ok'])

        job_id = self.drac_client.create_config_job(
            uris.DCIM_BIOSService, cim_creation_class_name, cim_name, target,
            reboot=True, realtime=False)

        mock_invoke.assert_called_once_with(
            mock.ANY, uris.DCIM_BIOSService, 'CreateTargetedConfigJob',
            expected_selectors, expected_properties,
            expected_return_value=utils.RET_CREATED,
            wait_for_idrac=wait_for_idrac)
        self.assertEqual('JID_442507917525', job_id)

    @mock.patch.object(dracclient.client.WSManClient, 'invoke', spec_set=True,
                       autospec=True)
    def test_create_config_job_with_realtime(self, mock_invoke):
        cim_creation_class_name = 'DCIM_BIOSService'
        cim_name = 'DCIM:BIOSService'
        target = 'BIOS.Setup.1-1'
        wait_for_idrac = True
        expected_selectors = {'CreationClassName': cim_creation_class_name,
                              'Name': cim_name,
                              'SystemCreationClassName': 'DCIM_ComputerSystem',
                              'SystemName': 'DCIM:ComputerSystem'}
        expected_properties = {'Target': target,
                               'ScheduledStartTime': 'TIME_NOW',
                               'RealTime': '1'}
        mock_invoke.return_value = lxml.etree.fromstring(
            test_utils.JobInvocations[uris.DCIM_BIOSService][
                'CreateTargetedConfigJob']['ok'])

        job_id = self.drac_client.create_config_job(
            uris.DCIM_BIOSService, cim_creation_class_name, cim_name, target,
            reboot=False, realtime=True)

        mock_invoke.assert_called_once_with(
            mock.ANY, uris.DCIM_BIOSService, 'CreateTargetedConfigJob',
            expected_selectors, expected_properties,
            expected_return_value=utils.RET_CREATED,
            wait_for_idrac=wait_for_idrac)
        self.assertEqual('JID_442507917525', job_id)

    @mock.patch.object(dracclient.client.WSManClient, 'invoke', spec_set=True,
                       autospec=True)
    def test_create_reboot_job(self, mock_invoke):
        expected_selectors = {
            'SystemCreationClassName': 'DCIM_ComputerSystem',
            'SystemName': 'idrac',
            'CreationClassName': 'DCIM_JobService',
            'Name': 'JobService'}
        expected_properties = {'RebootJobType': '3'}
        self.drac_client.create_reboot_job()

        mock_invoke.assert_called_once_with(
            mock.ANY, uris.DCIM_JobService, 'CreateRebootJob',
            expected_selectors, expected_properties,
            expected_return_value=utils.RET_CREATED)

    def test_create_reboot_job_bad_type(self):
        self.assertRaises(
            exceptions.InvalidParameterValue,
            self.drac_client.create_reboot_job, 'BAD REBOOT TYPE')

    @mock.patch.object(dracclient.client.WSManClient, 'invoke', spec_set=True,
                       autospec=True)
    def test_delete_pending_config(self, mock_invoke):
        cim_creation_class_name = 'DCIM_BIOSService'
        cim_name = 'DCIM:BIOSService'
        target = 'BIOS.Setup.1-1'
        expected_selectors = {'CreationClassName': cim_creation_class_name,
                              'Name': cim_name,
                              'SystemCreationClassName': 'DCIM_ComputerSystem',
                              'SystemName': 'DCIM:ComputerSystem'}
        expected_properties = {'Target': target}
        mock_invoke.return_value = lxml.etree.fromstring(
            test_utils.JobInvocations[uris.DCIM_BIOSService][
                'DeletePendingConfiguration']['ok'])

        self.drac_client.delete_pending_config(
            uris.DCIM_BIOSService, cim_creation_class_name, cim_name, target)

        mock_invoke.assert_called_once_with(
            mock.ANY, uris.DCIM_BIOSService, 'DeletePendingConfiguration',
            expected_selectors, expected_properties,
            expected_return_value=utils.RET_SUCCESS)

    @requests_mock.Mocker()
    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    def test_delete_pending_config_failed(self, mock_requests,
                                          mock_wait_until_idrac_is_ready):
        cim_creation_class_name = 'DCIM_BIOSService'
        cim_name = 'DCIM:BIOSService'
        target = 'BIOS.Setup.1-1'
        mock_requests.post(
            'https://1.2.3.4:443/wsman',
            text=test_utils.JobInvocations[uris.DCIM_BIOSService][
                'DeletePendingConfiguration']['error'])

        self.assertRaises(
            exceptions.DRACOperationFailed,
            self.drac_client.delete_pending_config, uris.DCIM_BIOSService,
            cim_creation_class_name, cim_name, target)


class ClientJobScheduleTestCase(base.BaseTest):
    def setUp(self):
        super(ClientJobScheduleTestCase, self).setUp()
        self.drac_client = dracclient.client.DRACClient(
            **test_utils.FAKE_ENDPOINT)

    def _test_schedule_job_execution(self,
                                     mock_invoke,
                                     job_ids,
                                     start_time,
                                     expected_properties):
        expected_selectors = {
            'SystemCreationClassName': 'DCIM_ComputerSystem',
            'SystemName': 'idrac',
            'CreationClassName': 'DCIM_JobService',
            'Name': 'JobService'}

        if start_time is None:
            self.drac_client.schedule_job_execution(job_ids)
        else:
            self.drac_client.schedule_job_execution(job_ids, start_time)

        mock_invoke.assert_called_once_with(
            mock.ANY, uris.DCIM_JobService, 'SetupJobQueue',
            expected_selectors, expected_properties,
            expected_return_value=utils.RET_SUCCESS)

    @mock.patch.object(dracclient.client.WSManClient, 'invoke', spec_set=True,
                       autospec=True)
    def test_schedule_job_execution_one_job(self, mock_invoke):
        job_ids = ['JID_442507917525']
        expected_properties = {'StartTimeInterval': 'TIME_NOW',
                               'JobArray': job_ids}

        self._test_schedule_job_execution(mock_invoke, job_ids, None,
                                          expected_properties)

    @mock.patch.object(dracclient.client.WSManClient, 'invoke', spec_set=True,
                       autospec=True)
    def test_schedule_job_execution_multi_job(self, mock_invoke):
        job_ids = ['JID_442507917525', 'JID_442507917526']
        expected_properties = {'StartTimeInterval': 'TIME_NOW',
                               'JobArray': job_ids}
        self._test_schedule_job_execution(mock_invoke, job_ids, None,
                                          expected_properties)

    @mock.patch.object(dracclient.client.WSManClient, 'invoke', spec_set=True,
                       autospec=True)
    def test_schedule_job_execution_one_job_with_time(self, mock_invoke):
        job_ids = ['JID_442507917525']
        timestamp = datetime.datetime.today().strftime('%Y%m%d%H%M%S')
        expected_properties = {'StartTimeInterval': timestamp,
                               'JobArray': job_ids}
        self._test_schedule_job_execution(mock_invoke, job_ids, timestamp,
                                          expected_properties)

    @mock.patch.object(dracclient.client.WSManClient, 'invoke', spec_set=True,
                       autospec=True)
    def test_schedule_job_execution_multi_job_with_time(self, mock_invoke):
        job_ids = ['JID_442507917525', 'JID_442507917526']
        timestamp = datetime.datetime.today().strftime('%Y%m%d%H%M%S')
        expected_properties = {'StartTimeInterval': timestamp,
                               'JobArray': job_ids}
        self._test_schedule_job_execution(mock_invoke, job_ids, timestamp,
                                          expected_properties)

    @mock.patch.object(dracclient.client.WSManClient, 'invoke')
    def test_schedule_job_execution_no_jobs(self, mock_invoke):
        self.drac_client.schedule_job_execution(job_ids=[])
        mock_invoke.assert_not_called()



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import collections
import lxml.etree
import random
import re
from unittest import mock

import requests_mock

import dracclient.client
from dracclient import constants
from dracclient import exceptions
import dracclient.resources.job
from dracclient.resources import raid
from dracclient.resources import uris
from dracclient.tests import base
from dracclient.tests import utils as test_utils
from dracclient import utils


@requests_mock.Mocker()
class ClientRAIDManagementTestCase(base.BaseTest):

    def setUp(self):
        super(ClientRAIDManagementTestCase, self).setUp()
        self.drac_client = dracclient.client.DRACClient(
            **test_utils.FAKE_ENDPOINT)
        self.raid_controller_fqdd = "RAID.Integrated.1-1"
        self.boss_controller_fqdd = "AHCI.Slot.3-1"
        cntl_dict = {'RAID.Integrated.1-1':
                     ['Disk.Bay.0:Enclosure.Internal.0-1:RAID.Integrated.1-1',
                      'Disk.Bay.1:Enclosure.Internal.0-1:RAID.Integrated.1-1'],
                     'AHCI.Integrated.1-1':
                     ['Disk.Bay.0:Enclosure.Internal.0-1:AHCI.Integrated.1-1',
                      'Disk.Bay.1:Enclosure.Internal.0-1:AHCI.Integrated.1-1']}
        self.controllers_to_physical_disk_ids = cntl_dict
        self.disk_1 = raid.PhysicalDisk(
            id='Disk.Bay.0:Enclosure.Internal.0-1:RAID.Integrated.1-1',
            description='Disk 0 in Backplane 1 of Int RAID Controller 1',
            controller='RAID.Integrated.1-1',
            manufacturer='ATA',
            model='ST91000640NS',
            media_type='hdd',
            interface_type='sata',
            size_mb=953344,
            free_size_mb=953344,
            serial_number='9XG4SLGZ',
            firmware_version='AA09',
            status='ok',
            raid_status='ready',
            sas_address='500056B37789ABE3',
            device_protocol=None,
            bus=None)

        self.disk_2 = raid.PhysicalDisk(
            id='Disk.Bay.1:Enclosure.Internal.0-1:RAID.Integrated.1-1',
            description='Disk 1 in Backplane 1 of Int RAID Controller 1',
            controller='RAID.Integrated.1-1',
            manufacturer='ATA',
            model='ST91000640NS',
            media_type='hdd',
            interface_type='sata',
            size_mb=953344,
            free_size_mb=953344,
            serial_number='9XG4SLGZ',
            firmware_version='AA09',
            status='online',
            raid_status='ready',
            sas_address='500056B37789ABE3',
            device_protocol=None,
            bus=None)

        self.disk_3 = raid.PhysicalDisk(
            id='Disk.Bay.0:Enclosure.Internal.0-1:AHCI.Integrated.1-1',
            description='Disk 1 in Backplane 1 of Int BOSS Controller 1',
            controller='AHCI.Integrated.1-1',
            manufacturer='ATA',
            model='ST91000640NS',
            media_type='hdd',
            interface_type='sata',
            size_mb=953344,
            free_size_mb=953344,
            serial_number='9XG4SLGZ',
            firmware_version='AA09',
            status='online',
            raid_status='ready',
            sas_address='500056B37789ABE3',
            device_protocol=None,
            bus=None)

        self.disk_4 = raid.PhysicalDisk(
            id='Disk.Bay.1:Enclosure.Internal.0-1:AHCI.Integrated.1-1',
            description='Disk 1 in Backplane 1 of Int RAID Controller 1',
            controller='AHCI.Integrated.1-1',
            manufacturer='ATA',
            model='ST91000640NS',
            media_type='hdd',
            interface_type='sata',
            size_mb=953344,
            free_size_mb=953344,
            serial_number='9XG4SLGZ',
            firmware_version='AA09',
            status='online',
            raid_status='ready',
            sas_address='500056B37789ABE3',
            device_protocol=None,
            bus=None)

    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    def test_list_raid_settings(self, mock_requests,
                                mock_wait_until_idrac_is_ready):
        expected_enum_attr = raid.RAIDEnumerableAttribute(
            name='RAIDCurrentControllerMode',
            instance_id='RAID.Integrated.1-1:RAIDCurrentControllerMode',  # noqa
            current_value=['RAID'],
            pending_value=None,
            read_only=True,
            fqdd='RAID.Integrated.1-1',
            possible_values=['RAID', 'Enhanced HBA'])
        expected_string_attr = raid.RAIDStringAttribute(
            name='Name',
            instance_id='Disk.Virtual.1:RAID.Integrated.1-1:Name',  # noqa
            current_value='Virtual Disk 1',
            pending_value=None,
            read_only=True,
            fqdd='Disk.Virtual.1:RAID.Integrated.1-1',
            min_length=0,
            max_length=129)
        expected_integer_attr = raid.RAIDIntegerAttribute(
            name='RAIDmaxSupportedVD',
            instance_id='RAID.Integrated.1-1:RAIDmaxSupportedVD',  # noqa
            current_value=240,
            pending_value=None,
            read_only=True,
            fqdd='RAID.Integrated.1-1',
            lower_bound=0,
            upper_bound=0)
        # expected_string_attr
        mock_requests.post('https://1.2.3.4:443/wsman', [
            {'text': test_utils.RAIDEnumerations[
                uris.DCIM_RAIDEnumeration]['ok']},
            {'text': test_utils.RAIDEnumerations[
                uris.DCIM_RAIDString]['ok']},
            {'text': test_utils.RAIDEnumerations[
                uris.DCIM_RAIDInteger]['ok']}
        ])

        raid_settings = self.drac_client.list_raid_settings()
        self.assertEqual(219, len(raid_settings))
        # enumerable attribute
        self.assertIn(
            'RAID.Integrated.1-1:RAIDCurrentControllerMode',  # noqa
            raid_settings)
        self.assertEqual(expected_enum_attr.fqdd, raid_settings[
                         'RAID.Integrated.1-1:RAIDCurrentControllerMode'].fqdd)  # noqa
        # string attribute
        self.assertIn(
            'Disk.Virtual.1:RAID.Integrated.1-1:Name',  # noqa
            raid_settings)
        self.assertEqual(expected_string_attr.fqdd,
                         raid_settings['Disk.Virtual.1:RAID.Integrated.1-1:Name'].fqdd)  # noqa
        # integer attribute
        self.assertIn(
            'RAID.Integrated.1-1:RAIDmaxSupportedVD',  # noqa
            raid_settings)
        self.assertEqual(expected_integer_attr.fqdd, raid_settings[
            'RAID.Integrated.1-1:RAIDmaxSupportedVD'].fqdd)  # noqa

    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    @mock.patch.object(dracclient.client.WSManClient,
                       'invoke', spec_set=True,
                       autospec=True)
    def test_set_raid_settings(self, mock_requests,
                               mock_invoke,
                               mock_wait_until_idrac_is_ready):
        mock_requests.post('https://1.2.3.4:443/wsman', [
            {'text': test_utils.RAIDEnumerations[
                uris.DCIM_RAIDEnumeration]['ok']},
            {'text': test_utils.RAIDEnumerations[
                uris.DCIM_RAIDString]['ok']},
            {'text': test_utils.RAIDEnumerations[
                uris.DCIM_RAIDInteger]['ok']}])
        mock_invoke.return_value = lxml.etree.fromstring(
            test_utils.RAIDInvocations[uris.DCIM_RAIDService][
                'SetAttributes']['ok'])

        result = self.drac_client.set_raid_settings(
            self.raid_controller_fqdd,
            {'RAID.Integrated.1-1:RAIDRequestedControllerMode': 'RAID'})

        self.assertEqual({'is_commit_required': True,
                          'is_reboot_required': constants.RebootRequired.true
                          },
                         result)

    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    def test_set_raid_settings_with_unknown_attr(
            self, mock_requests, mock_wait_until_idrac_is_ready):
        mock_requests.post('https://1.2.3.4:443/wsman', [
            {'text': test_utils.RAIDEnumerations[
                uris.DCIM_RAIDEnumeration]['ok']},
            {'text': test_utils.RAIDEnumerations[
                uris.DCIM_RAIDString]['ok']},
            {'text': test_utils.RAIDEnumerations[
                uris.DCIM_RAIDInteger]['ok']},
            {'text': test_utils.RAIDInvocations[
                uris.DCIM_RAIDService]['SetAttributes']['error']}])

        self.assertRaises(exceptions.InvalidParameterValue,
                          self.drac_client.set_raid_settings,
                          self.raid_controller_fqdd, {'foo': 'bar'})

    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    def test_set_raid_settings_with_unchanged_attr(
            self, mock_requests, mock_wait_until_idrac_is_ready):
        mock_requests.post('https://1.2.3.4:443/wsman', [
            {'text': test_utils.RAIDEnumerations[
                uris.DCIM_RAIDEnumeration]['ok']},
            {'text': test_utils.RAIDEnumerations[
                uris.DCIM_RAIDString]['ok']},
            {'text': test_utils.RAIDEnumerations[
                uris.DCIM_RAIDInteger]['ok']}])
        attrKey = 'Disk.Virtual.1:RAID.Integrated.1-1:RAIDdefaultWritePolicy'
        result = self.drac_client.set_raid_settings(
            self.raid_controller_fqdd,
            {attrKey: 'WriteBack'})

        self.assertEqual({'is_commit_required': False,
                          'is_reboot_required':
                          constants.RebootRequired.false},
                         result)

    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    def test_set_raid_settings_with_readonly_attr(
            self, mock_requests, mock_wait_until_idrac_is_ready):
        expected_message = (
            "Cannot set read-only RAID attributes: "
            "['RAID.Integrated.1-1:RAIDCurrentControllerMode']."
        )
        mock_requests.post('https://1.2.3.4:443/wsman', [
            {'text': test_utils.RAIDEnumerations[
                uris.DCIM_RAIDEnumeration]['ok']},
            {'text': test_utils.RAIDEnumerations[
                uris.DCIM_RAIDString]['ok']},
            {'text': test_utils.RAIDEnumerations[
                uris.DCIM_RAIDInteger]['ok']}])

        self.assertRaisesRegexp(
            exceptions.DRACOperationFailed, re.escape(expected_message),
            self.drac_client.set_raid_settings,
            self.raid_controller_fqdd,
            {'RAID.Integrated.1-1:RAIDCurrentControllerMode': 'Enhanced HBA'})

    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    def test_set_raid_settings_with_incorrect_enum_value(
            self, mock_requests, mock_wait_until_idrac_is_ready):
        expected_message = ("Attribute 'RAIDRequestedControllerMode' cannot "
                            "be set to value 'foo'. It must be in "
                            "['RAID', 'Enhanced HBA', 'None'].")

        mock_requests.post('https://1.2.3.4:443/wsman', [
            {'text': test_utils.RAIDEnumerations[
                uris.DCIM_RAIDEnumeration]['ok']},
            {'text': test_utils.RAIDEnumerations[
                uris.DCIM_RAIDString]['ok']},
            {'text': test_utils.RAIDEnumerations[
                uris.DCIM_RAIDInteger]['ok']}])
        self.assertRaisesRegexp(
            exceptions.DRACOperationFailed, re.escape(expected_message),
            self.drac_client.set_raid_settings,
            self.raid_controller_fqdd,
            {'RAID.Integrated.1-1:RAIDRequestedControllerMode': 'foo'})

    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    def test_list_raid_controllers(self, mock_requests,
                                   mock_wait_until_idrac_is_ready):
        expected_raid_controller = raid.RAIDController(
            id='RAID.Integrated.1-1',
            description='Integrated RAID Controller 1',
            manufacturer='DELL',
            model='PERC H710 Mini',
            primary_status='ok',
            firmware_version='21.3.0-0009',
            bus='1',
            supports_realtime=True)
        mock_requests.post(
            'https://1.2.3.4:443/wsman',
            text=test_utils.RAIDEnumerations[uris.DCIM_ControllerView]['ok'])

        self.assertIn(expected_raid_controller,
                      self.drac_client.list_raid_controllers())

    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    def test_list_virtual_disks(self, mock_requests,
                                mock_wait_until_idrac_is_ready):
        expected_virtual_disk = raid.VirtualDisk(
            id='Disk.Virtual.0:RAID.Integrated.1-1',
            name='disk 0',
            description='Virtual Disk 0 on Integrated RAID Controller 1',
            controller='RAID.Integrated.1-1',
            raid_level='1',
            size_mb=571776,
            status='ok',
            raid_status='online',
            span_depth=1,
            span_length=2,
            pending_operations=None,
            physical_disks=[
                'Disk.Bay.0:Enclosure.Internal.0-1:RAID.Integrated.1-1',
                'Disk.Bay.1:Enclosure.Internal.0-1:RAID.Integrated.1-1'
            ])

        mock_requests.post(
            'https://1.2.3.4:443/wsman',
            text=test_utils.RAIDEnumerations[uris.DCIM_VirtualDiskView]['ok'])

        self.assertIn(expected_virtual_disk,
                      self.drac_client.list_virtual_disks())

    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    def test_list_virtual_disks_with_raid_status_change(
            self, mock_requests, mock_wait_until_idrac_is_ready):
        expected_virtual_disk = raid.VirtualDisk(
            id='Disk.Virtual.0:RAID.Integrated.1-1',
            name='disk 0',
            description='Virtual Disk 0 on Integrated RAID Controller 1',
            controller='RAID.Integrated.1-1',
            raid_level='1',
            size_mb=571776,
            status='ok',
            raid_status='online',
            span_depth=1,
            span_length=2,
            pending_operations=None,
            physical_disks=[
                'Disk.Bay.4:Enclosure.Internal.0-1:RAID.Integrated.1-1',
                'Disk.Bay.5:Enclosure.Internal.0-1:RAID.Integrated.1-1'
            ])

        mock_requests.post(
            'https://1.2.3.4:443/wsman',
            text=test_utils.RAIDEnumerations[
                uris.DCIM_VirtualDiskView]['Raid_Status_ok'])

        self.assertIn(expected_virtual_disk,
                      self.drac_client.list_virtual_disks())

    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    def test_list_physical_disks(self, mock_requests,
                                 mock_wait_until_idrac_is_ready):
        expected_physical_disk = raid.PhysicalDisk(
            id='Disk.Bay.1:Enclosure.Internal.0-1:RAID.Integrated.1-1',
            description=('Disk 1 in Backplane 1 of '
                         'Integrated RAID Controller 1'),
            controller='RAID.Integrated.1-1',
            manufacturer='SEAGATE',
            model='ST600MM0006',
            media_type='hdd',
            interface_type='sas',
            size_mb=571776,
            free_size_mb=571776,
            serial_number='S0M3EY2Z',
            firmware_version='LS0A',
            status='ok',
            raid_status='ready',
            sas_address='5000C5007764F409',
            device_protocol=None,
            bus=None)

        mock_requests.post(
            'https://1.2.3.4:443/wsman',
            text=test_utils.RAIDEnumerations[uris.DCIM_PhysicalDiskView]['ok'])

        self.assertIn(expected_physical_disk,
                      self.drac_client.list_physical_disks())

    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    def test_list_physical_disks_direct(self, mock_requests,
                                        mock_wait_until_idrac_is_ready):
        expected_physical_disk = raid.PhysicalDisk(
            id='Disk.Direct.2:RAID.Integrated.1-1',
            description=('Disk 2 on '
                         'Integrated RAID Controller 1'),
            controller='RAID.Integrated.1-1',
            manufacturer='ATA',
            model='ST600MM0007',
            media_type='ssd',
            interface_type='sata',
            size_mb=571776,
            free_size_mb=571776,
            serial_number='S0M3EY3Z',
            firmware_version='LS0B',
            status='ok',
            raid_status='ready',
            sas_address='5000C5007764F409',
            device_protocol=None,
            bus=None)

        mock_requests.post(
            'https://1.2.3.4:443/wsman',
            text=test_utils.RAIDEnumerations[uris.DCIM_PhysicalDiskView]['ok'])

        self.assertIn(expected_physical_disk,
                      self.drac_client.list_physical_disks())

    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    def test_list_physical_disks_nvme(self, mock_requests,
                                      mock_wait_until_idrac_is_ready):
        expected_physical_disk = raid.PhysicalDisk(
            id='Disk.Bay.20:Enclosure.Internal.0-1:PCIeExtender.Slot.1',
            description='PCIe SSD in Slot 20 in Bay 1',
            controller='PCIeExtender.Slot.1',
            manufacturer='SAMSUNG',
            model='Dell Express Flash PM1725a 800GB SFF',
            media_type='ssd',
            interface_type='pcie',
            size_mb=763097,
            free_size_mb=None,
            serial_number='S39YNX0JB02343',
            firmware_version='1.0.4',
            status='unknown',
            raid_status=None,
            sas_address=None,
            device_protocol='NVMe-MI1.0',
            bus='3E')

        mock_requests.post(
            'https://1.2.3.4:443/wsman',
            text=test_utils.RAIDEnumerations[uris.DCIM_PhysicalDiskView]['ok'])

        self.assertIn(expected_physical_disk,
                      self.drac_client.list_physical_disks())

    # Verify that various client convert_physical_disks calls to dracclient
    # result in a WSMan.invoke with appropriate parameters
    def _random_term(self):
        return "".join(random.sample('ABCDEFGHabcdefgh0123456',
                                     random.randint(4, 12)))

    def _random_fqdd(self):
        result = self._random_term()
        for i in range(0, random.randint(6, 10)):
            result += random.sample('.:-', 1)[0] + self._random_term()
        return result

    @mock.patch.object(dracclient.client.WSManClient, 'invoke',
                       spec_set=True, autospec=True)
    def test_convert_physical_disks_1(self, mock_requests, mock_invoke):
        '''Convert a single disk to RAID mode'''
        device_fqdd = self._random_fqdd()
        expected_invocation = 'ConvertToRAID'
        expected_selectors = {'SystemCreationClassName': 'DCIM_ComputerSystem',
                              'CreationClassName': 'DCIM_RAIDService',
                              'SystemName': 'DCIM:ComputerSystem',
                              'Name': 'DCIM:RAIDService'}
        expected_properties = {'PDArray': [device_fqdd]}
        mock_invoke.return_value = lxml.etree.fromstring(
            test_utils.RAIDInvocations[uris.DCIM_RAIDService][
                'ConvertToRAID']['ok'])

        result = self.drac_client.convert_physical_disks(
            raid_controller='controller',
            physical_disks=[device_fqdd],
            raid_enable=True)

        self.assertEqual({'is_commit_required': True,
                          'is_reboot_required': constants.RebootRequired.true},
                         result)
        mock_invoke.assert_called_once_with(
            mock.ANY, uris.DCIM_RAIDService, expected_invocation,
            expected_selectors, expected_properties,
            expected_return_value=utils.RET_SUCCESS)

    @mock.patch.object(dracclient.client.WSManClient, 'invoke',
                       spec_set=True, autospec=True)
    def test_convert_physical_disks_n(self, mock_requests, mock_invoke):
        '''Convert a number of disks to RAID mode'''
        device_list = []
        for i in range(0, random.randint(2, 10)):
            device_list += self._random_fqdd()

        expected_invocation = 'ConvertToRAID'
        expected_selectors = {'SystemCreationClassName': 'DCIM_ComputerSystem',
                              'CreationClassName': 'DCIM_RAIDService',
                              'SystemName': 'DCIM:ComputerSystem',
                              'Name': 'DCIM:RAIDService'}
        expected_properties = {'PDArray': device_list}
        mock_invoke.return_value = lxml.etree.fromstring(
            test_utils.RAIDInvocations[uris.DCIM_RAIDService][
                'ConvertToRAID']['ok'])

        result = self.drac_client.convert_physical_disks(
            raid_controller='controller',
            physical_disks=device_list,
            raid_enable=True)

        self.assertEqual({'is_commit_required': True,
                          'is_reboot_required': constants.RebootRequired.true},
                         result)
        mock_invoke.assert_called_once_with(
            mock.ANY, uris.DCIM_RAIDService, expected_invocation,
            expected_selectors, expected_properties,
            expected_return_value=utils.RET_SUCCESS)

    @mock.patch.object(dracclient.client.WSManClient, 'invoke',
                       spec_set=True, autospec=True)
    def test_convert_physical_disks_nonraid_1(self, mock_requests,
                                              mock_invoke):
        '''Convert a single disk to non-RAID mode'''
        device_fqdd = self._random_fqdd()
        expected_invocation = 'ConvertToNonRAID'
        expected_selectors = {'SystemCreationClassName': 'DCIM_ComputerSystem',
                              'CreationClassName': 'DCIM_RAIDService',
                              'SystemName': 'DCIM:ComputerSystem',
                              'Name': 'DCIM:RAIDService'}
        expected_properties = {'PDArray': [device_fqdd]}
        mock_invoke.return_value = lxml.etree.fromstring(
            test_utils.RAIDInvocations[uris.DCIM_RAIDService][
                'ConvertToRAID']['ok'])

        result = self.drac_client.convert_physical_disks(
            raid_controller='controller',
            physical_disks=[device_fqdd],
            raid_enable=False)

        self.assertEqual({'is_commit_required': True,
                          'is_reboot_required': constants.RebootRequired.true},
                         result)
        mock_invoke.assert_called_once_with(
            mock.ANY, uris.DCIM_RAIDService, expected_invocation,
            expected_selectors, expected_properties,
            expected_return_value=utils.RET_SUCCESS)

    @mock.patch.object(dracclient.client.WSManClient, 'invoke',
                       spec_set=True, autospec=True)
    def test_convert_physical_disks_nonraid_n(self, mock_requests,
                                              mock_invoke):
        '''Convert a number of disks to non-RAID mode'''
        device_list = []
        for i in range(0, random.randint(2, 10)):
            device_list += self._random_fqdd()

        expected_invocation = 'ConvertToNonRAID'
        expected_selectors = {'SystemCreationClassName': 'DCIM_ComputerSystem',
                              'CreationClassName': 'DCIM_RAIDService',
                              'SystemName': 'DCIM:ComputerSystem',
                              'Name': 'DCIM:RAIDService'}
        expected_properties = {'PDArray': device_list}
        mock_invoke.return_value = lxml.etree.fromstring(
            test_utils.RAIDInvocations[uris.DCIM_RAIDService][
                'ConvertToRAID']['ok'])

        result = self.drac_client.convert_physical_disks(
            raid_controller='controller',
            physical_disks=device_list,
            raid_enable=False)

        self.assertEqual({'is_commit_required': True,
                          'is_reboot_required': constants.RebootRequired.true},
                         result)
        mock_invoke.assert_called_once_with(
            mock.ANY, uris.DCIM_RAIDService, expected_invocation,
            expected_selectors, expected_properties,
            expected_return_value=utils.RET_SUCCESS)

    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    def test_convert_physical_disks_fail(self, mock_requests,
                                         mock_wait_until_idrac_is_ready):
        mock_requests.post(
            'https://1.2.3.4:443/wsman',
            text=test_utils.RAIDInvocations[
                uris.DCIM_RAIDService]['ConvertToRAID']['error'])

        self.assertRaises(
            exceptions.DRACOperationFailed,
            self.drac_client.convert_physical_disks,
            raid_controller='controller',
            physical_disks=['Disk0:Enclosure-1:RAID-1',
                            'Disk1:Enclosure-1:RAID-1'],
            raid_enable=True)

    @mock.patch.object(dracclient.client.WSManClient, 'invoke',
                       spec_set=True, autospec=True)
    def test_create_virtual_disk(self, mock_requests, mock_invoke):
        expected_selectors = {'SystemCreationClassName': 'DCIM_ComputerSystem',
                              'CreationClassName': 'DCIM_RAIDService',
                              'SystemName': 'DCIM:ComputerSystem',
                              'Name': 'DCIM:RAIDService'}
        expected_properties = {'Target': 'controller',
                               'PDArray': ['disk1', 'disk2'],
                               'VDPropNameArray': ['Size', 'RAIDLevel'],
                               'VDPropValueArray': ['42', '4']}
        mock_invoke.return_value = lxml.etree.fromstring(
            test_utils.RAIDInvocations[uris.DCIM_RAIDService][
                'CreateVirtualDisk']['ok'])

        result = self.drac_client.create_virtual_disk(
            raid_controller='controller', physical_disks=['disk1', 'disk2'],
            raid_level='1', size_mb=42)

        self.assertEqual({'is_commit_required': True,
                          'is_reboot_required': constants.RebootRequired.true},
                         result)
        mock_invoke.assert_called_once_with(
            mock.ANY, uris.DCIM_RAIDService, 'CreateVirtualDisk',
            expected_selectors, expected_properties,
            expected_return_value=utils.RET_SUCCESS)

    @mock.patch.object(dracclient.client.WSManClient, 'invoke',
                       spec_set=True, autospec=True)
    def test_create_virtual_disk_with_extra_params(self, mock_requests,
                                                   mock_invoke):
        expected_selectors = {'SystemCreationClassName': 'DCIM_ComputerSystem',
                              'CreationClassName': 'DCIM_RAIDService',
                              'SystemName': 'DCIM:ComputerSystem',
                              'Name': 'DCIM:RAIDService'}
        expected_properties = {'Target': 'controller',
                               'PDArray': ['disk1', 'disk2'],
                               'VDPropNameArray': ['Size', 'RAIDLevel',
                                                   'VirtualDiskName',
                                                   'SpanDepth', 'SpanLength'],
                               'VDPropValueArray': ['42', '4', 'name', '3',
                                                    '2']}
        mock_invoke.return_value = lxml.etree.fromstring(
            test_utils.RAIDInvocations[uris.DCIM_RAIDService][
                'CreateVirtualDisk']['ok'])

        result = self.drac_client.create_virtual_disk(
            raid_controller='controller', physical_disks=['disk1', 'disk2'],
            raid_level='1', size_mb=42, disk_name='name', span_length=2,
            span_depth=3)

        self.assertEqual({'is_commit_required': True,
                          'is_reboot_required': constants.RebootRequired.true},
                         result)
        mock_invoke.assert_called_once_with(
            mock.ANY, uris.DCIM_RAIDService, 'CreateVirtualDisk',
            expected_selectors, expected_properties,
            expected_return_value=utils.RET_SUCCESS)

    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    def test_create_virtual_disk_fail(self, mock_requests,
                                      mock_wait_until_idrac_is_ready):
        mock_requests.post(
            'https://1.2.3.4:443/wsman',
            text=test_utils.RAIDInvocations[
                uris.DCIM_RAIDService]['CreateVirtualDisk']['error'])

        self.assertRaises(
            exceptions.DRACOperationFailed,
            self.drac_client.create_virtual_disk, raid_controller='controller',
            physical_disks=['disk1', 'disk2'], raid_level='1', size_mb=42,
            disk_name='name', span_length=2, span_depth=3)

    def test_create_virtual_disk_missing_controller(self, mock_requests):
        self.assertRaises(
            exceptions.InvalidParameterValue,
            self.drac_client.create_virtual_disk, raid_controller=None,
            physical_disks=['disk1', 'disk2'], raid_level='1', size_mb=42,
            disk_name='name', span_length=2, span_depth=3)

    def test_create_virtual_disk_missing_physical_disks(self, mock_requests):
        self.assertRaises(
            exceptions.InvalidParameterValue,
            self.drac_client.create_virtual_disk, raid_controller='controller',
            physical_disks=None, raid_level='1', size_mb=42,
            disk_name='name', span_length=2, span_depth=3)

    def test_create_virtual_disk_missing_raid_level(self, mock_requests):
        self.assertRaises(
            exceptions.InvalidParameterValue,
            self.drac_client.create_virtual_disk, raid_controller='controller',
            physical_disks=['disk1', 'disk2'], raid_level=None, size_mb=42,
            disk_name='name', span_length=2, span_depth=3)

    def test_create_virtual_disk_invalid_raid_level(self, mock_requests):
        self.assertRaises(
            exceptions.InvalidParameterValue,
            self.drac_client.create_virtual_disk, raid_controller='controller',
            physical_disks=['disk1', 'disk2'], raid_level='foo', size_mb=42,
            disk_name='name', span_length=2, span_depth=3)

    def test_create_virtual_disk_missing_size(self, mock_requests):
        self.assertRaises(
            exceptions.InvalidParameterValue,
            self.drac_client.create_virtual_disk, raid_controller='controller',
            physical_disks=['disk1', 'disk2'], raid_level='1', size_mb=None,
            disk_name='name', span_length=2, span_depth=3)

    # This test is specifically for support of creating a RAID1 on a Dell BOSS
    # card.  It requires that size_mb is set to 0
    @mock.patch.object(dracclient.client.WSManClient, 'invoke',
                       spec_set=True, autospec=True)
    def test_create_virtual_disk_0_size(self, mock_requests, mock_invoke):
        mock_invoke.return_value = lxml.etree.fromstring(
            test_utils.RAIDInvocations[uris.DCIM_RAIDService][
                'CreateVirtualDisk']['ok'])

        self.drac_client.create_virtual_disk(
            raid_controller='controller',
            physical_disks=['disk1', 'disk2'],
            raid_level='1',
            size_mb=0,
            disk_name='name',
            span_length=1,
            span_depth=2)

    def test_create_virtual_disk_invalid_size(self, mock_requests):
        self.assertRaises(
            exceptions.InvalidParameterValue,
            self.drac_client.create_virtual_disk, raid_controller='controller',
            physical_disks=['disk1', 'disk2'], raid_level='1',
            size_mb='foo', disk_name='name', span_length=2, span_depth=3)

    def test_create_virtual_disk_invalid_span_length(self, mock_requests):
        self.assertRaises(
            exceptions.InvalidParameterValue,
            self.drac_client.create_virtual_disk, raid_controller='controller',
            physical_disks=['disk1', 'disk2'], raid_level='1', size_mb=42,
            disk_name='name', span_length='foo', span_depth=3)

    def test_create_virtual_disk_invalid_span_depth(self, mock_requests):
        self.assertRaises(
            exceptions.InvalidParameterValue,
            self.drac_client.create_virtual_disk, raid_controller='controller',
            physical_disks=['disk1', 'disk2'], raid_level='1', size_mb=42,
            disk_name='name', span_length=2, span_depth='foo')

    @mock.patch.object(dracclient.client.WSManClient, 'invoke',
                       spec_set=True, autospec=True)
    def test_delete_virtual_disk(self, mock_requests, mock_invoke):
        expected_selectors = {'SystemCreationClassName': 'DCIM_ComputerSystem',
                              'CreationClassName': 'DCIM_RAIDService',
                              'SystemName': 'DCIM:ComputerSystem',
                              'Name': 'DCIM:RAIDService'}
        expected_properties = {'Target': 'disk1'}
        mock_invoke.return_value = lxml.etree.fromstring(
            test_utils.RAIDInvocations[uris.DCIM_RAIDService][
                'DeleteVirtualDisk']['ok'])

        result = self.drac_client.delete_virtual_disk('disk1')

        self.assertEqual({'is_commit_required': True,
                          'is_reboot_required': constants.RebootRequired.true},
                         result)
        mock_invoke.assert_called_once_with(
            mock.ANY, uris.DCIM_RAIDService, 'DeleteVirtualDisk',
            expected_selectors, expected_properties,
            expected_return_value=utils.RET_SUCCESS)

    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    def test_delete_virtual_disk_fail(self, mock_requests,
                                      mock_wait_until_idrac_is_ready):
        mock_requests.post(
            'https://1.2.3.4:443/wsman',
            text=test_utils.RAIDInvocations[
                uris.DCIM_RAIDService]['DeleteVirtualDisk']['error'])

        self.assertRaises(
            exceptions.DRACOperationFailed,
            self.drac_client.delete_virtual_disk, 'disk1')

    @mock.patch.object(dracclient.client.WSManClient, 'invoke',
                       spec_set=True, autospec=True)
    def test_reset_raid_config(self, mock_requests, mock_invoke):
        expected_selectors = {'SystemCreationClassName': 'DCIM_ComputerSystem',
                              'CreationClassName': 'DCIM_RAIDService',
                              'SystemName': 'DCIM:ComputerSystem',
                              'Name': 'DCIM:RAIDService'}
        expected_properties = {'Target': self.raid_controller_fqdd}
        mock_invoke.return_value = lxml.etree.fromstring(
            test_utils.RAIDInvocations[uris.DCIM_RAIDService][
                'ResetConfig']['ok'])
        result = self.drac_client.reset_raid_config(self.raid_controller_fqdd)
        self.assertEqual({'is_commit_required': True,
                          'is_reboot_required':
                          constants.RebootRequired.optional},
                         result)
        mock_invoke.assert_called_once_with(
            mock.ANY, uris.DCIM_RAIDService, 'ResetConfig',
            expected_selectors, expected_properties,
            expected_return_value=utils.RET_SUCCESS)

    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    def test_reset_raid_config_fail(self, mock_requests,
                                    mock_wait_until_idrac_is_ready):
        mock_requests.post(
            'https://1.2.3.4:443/wsman',
            text=test_utils.RAIDInvocations[
                uris.DCIM_RAIDService]['ResetConfig']['error'])

        self.assertRaises(
            exceptions.DRACOperationFailed,
            self.drac_client.reset_raid_config, self.raid_controller_fqdd)

    @mock.patch.object(dracclient.client.WSManClient, 'invoke',
                       spec_set=True, autospec=True)
    def test_clear_foreign_config(self, mock_requests, mock_invoke):
        expected_selectors = {'SystemCreationClassName': 'DCIM_ComputerSystem',
                              'CreationClassName': 'DCIM_RAIDService',
                              'SystemName': 'DCIM:ComputerSystem',
                              'Name': 'DCIM:RAIDService'}
        expected_properties = {'Target': self.raid_controller_fqdd}
        mock_invoke.return_value = lxml.etree.fromstring(
            test_utils.RAIDInvocations[uris.DCIM_RAIDService][
                'ClearForeignConfig']['ok'])

        result = self.drac_client.clear_foreign_config(
                self.raid_controller_fqdd)
        self.assertEqual({'is_commit_required': True,
                          'is_reboot_required':
                          constants.RebootRequired.optional},
                         result)
        mock_invoke.assert_called_once_with(
            mock.ANY, uris.DCIM_RAIDService, 'ClearForeignConfig',
            expected_selectors, expected_properties,
            check_return_value=False)

    @mock.patch.object(dracclient.client.WSManClient, 'invoke',
                       spec_set=True, autospec=True)
    def test_clear_foreign_config_with_no_foreign_drive(self,
                                                        mock_requests,
                                                        mock_invoke):
        expected_selectors = {'SystemCreationClassName': 'DCIM_ComputerSystem',
                              'CreationClassName': 'DCIM_RAIDService',
                              'SystemName': 'DCIM:ComputerSystem',
                              'Name': 'DCIM:RAIDService'}
        expected_properties = {'Target': self.raid_controller_fqdd}
        mock_invoke.return_value = lxml.etree.fromstring(
            test_utils.RAIDInvocations[uris.DCIM_RAIDService][
                'ClearForeignConfig']['no_foreign_drive'])

        result = self.drac_client.clear_foreign_config(
                self.raid_controller_fqdd)
        self.assertEqual({'is_commit_required': False,
                          'is_reboot_required':
                          constants.RebootRequired.false},
                         result)
        mock_invoke.assert_called_once_with(
            mock.ANY, uris.DCIM_RAIDService, 'ClearForeignConfig',
            expected_selectors, expected_properties,
            check_return_value=False)

    @mock.patch.object(dracclient.client.WSManClient, 'invoke',
                       spec_set=True, autospec=True)
    def test_clear_foreign_config_with_operation_not_supported(self,
                                                               mock_requests,
                                                               mock_invoke):
        expected_selectors = {'SystemCreationClassName': 'DCIM_ComputerSystem',
                              'CreationClassName': 'DCIM_RAIDService',
                              'SystemName': 'DCIM:ComputerSystem',
                              'Name': 'DCIM:RAIDService'}
        expected_properties = {'Target': self.boss_controller_fqdd}
        mock_invoke.return_value = lxml.etree.fromstring(
            test_utils.RAIDInvocations[uris.DCIM_RAIDService][
                'ClearForeignConfig']['foreign_drive_operation_not_supported'])

        result = self.drac_client.clear_foreign_config(
                self.boss_controller_fqdd)
        self.assertEqual({'is_commit_required': False,
                          'is_reboot_required':
                          constants.RebootRequired.false},
                         result)
        mock_invoke.assert_called_once_with(
            mock.ANY, uris.DCIM_RAIDService, 'ClearForeignConfig',
            expected_selectors, expected_properties,
            check_return_value=False)

    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    def test_clear_foreign_config_with_invalid_controller_id(
            self,
            mock_requests,
            mock_wait_until_idrac_is_ready):
        mock_requests.post(
            'https://1.2.3.4:443/wsman',
            text=test_utils.RAIDInvocations[
                uris.DCIM_RAIDService]['ClearForeignConfig']
            ['invalid_controller_id'])

        self.assertRaises(
            exceptions.DRACOperationFailed,
            self.drac_client.clear_foreign_config, 'bad')

    @mock.patch.object(dracclient.resources.job.JobManagement,
                       'create_config_job', spec_set=True, autospec=True)
    def test_commit_pending_raid_changes(self, mock_requests,
                                         mock_create_config_job):
        self.drac_client.commit_pending_raid_changes('controller',
                                                     realtime=False)

        mock_create_config_job.assert_called_once_with(
            mock.ANY, resource_uri=uris.DCIM_RAIDService,
            cim_creation_class_name='DCIM_RAIDService',
            cim_name='DCIM:RAIDService', target='controller', reboot=False,
            start_time='TIME_NOW', realtime=False)

    @mock.patch.object(dracclient.resources.job.JobManagement,
                       'create_config_job', spec_set=True, autospec=True)
    def test_commit_pending_raid_changes_with_reboot(self, mock_requests,
                                                     mock_create_config_job):
        self.drac_client.commit_pending_raid_changes('controller',
                                                     reboot=True,
                                                     realtime=False)

        mock_create_config_job.assert_called_once_with(
            mock.ANY, resource_uri=uris.DCIM_RAIDService,
            cim_creation_class_name='DCIM_RAIDService',
            cim_name='DCIM:RAIDService', target='controller', reboot=True,
            start_time='TIME_NOW', realtime=False)

    @mock.patch.object(dracclient.resources.job.JobManagement,
                       'create_config_job', spec_set=True, autospec=True)
    def test_commit_pending_raid_changes_with_start_time(
            self, mock_requests,
            mock_create_config_job):
        timestamp = '20140924140201'
        self.drac_client.commit_pending_raid_changes('controller',
                                                     start_time=timestamp,
                                                     realtime=False)

        mock_create_config_job.assert_called_once_with(
            mock.ANY, resource_uri=uris.DCIM_RAIDService,
            cim_creation_class_name='DCIM_RAIDService',
            cim_name='DCIM:RAIDService', target='controller', reboot=False,
            start_time=timestamp, realtime=False)

    @mock.patch.object(dracclient.resources.job.JobManagement,
                       'create_config_job', spec_set=True, autospec=True)
    def test_commit_pending_raid_changes_with_reboot_and_start_time(
            self, mock_requests,
            mock_create_config_job):
        timestamp = '20140924140201'
        self.drac_client.commit_pending_raid_changes('controller',
                                                     reboot=True,
                                                     start_time=timestamp,
                                                     realtime=False)

        mock_create_config_job.assert_called_once_with(
            mock.ANY, resource_uri=uris.DCIM_RAIDService,
            cim_creation_class_name='DCIM_RAIDService',
            cim_name='DCIM:RAIDService', target='controller', reboot=True,
            start_time=timestamp, realtime=False)

    @mock.patch.object(dracclient.resources.job.JobManagement,
                       'create_config_job', spec_set=True, autospec=True)
    def test_commit_pending_raid_changes_with_realtime(
            self, mock_requests,
            mock_create_config_job):
        timestamp = '20140924140201'
        self.drac_client.commit_pending_raid_changes('controller',
                                                     reboot=False,
                                                     start_time=timestamp,
                                                     realtime=True)

        mock_create_config_job.assert_called_once_with(
            mock.ANY, resource_uri=uris.DCIM_RAIDService,
            cim_creation_class_name='DCIM_RAIDService',
            cim_name='DCIM:RAIDService', target='controller', reboot=False,
            start_time=timestamp, realtime=True)

    @mock.patch.object(dracclient.resources.job.JobManagement,
                       'delete_pending_config', spec_set=True, autospec=True)
    def test_abandon_pending_raid_changes(self, mock_requests,
                                          mock_delete_pending_config):
        self.drac_client.abandon_pending_raid_changes('controller')

        mock_delete_pending_config.assert_called_once_with(
            mock.ANY, resource_uri=uris.DCIM_RAIDService,
            cim_creation_class_name='DCIM_RAIDService',
            cim_name='DCIM:RAIDService', target='controller')

    @mock.patch.object(dracclient.resources.job.JobManagement,
                       'delete_pending_config', spec_set=True, autospec=True)
    def test_abandon_pending_raid_changes_realtime(self, mock_requests,
                                                   mock_delete_pending_config):
        self.drac_client.abandon_pending_raid_changes('controller')

        mock_delete_pending_config.assert_called_once_with(
            mock.ANY, resource_uri=uris.DCIM_RAIDService,
            cim_creation_class_name='DCIM_RAIDService',
            cim_name='DCIM:RAIDService', target='controller')

    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    @mock.patch.object(dracclient.resources.raid.RAIDManagement,
                       'convert_physical_disks',
                       return_value={}, spec_set=True,
                       autospec=True)
    def test_raid_controller_jbod_capable(self, mock_requests,
                                          mock_wait_until_idrac_is_ready,
                                          mock_convert_physical_disks):

        mock_requests.post(
            'https://1.2.3.4:443/wsman',
            text=test_utils.RAIDEnumerations[uris.DCIM_PhysicalDiskView]['ok'])

        is_jbod = self.drac_client.is_jbod_capable(self.raid_controller_fqdd)

        self.assertTrue(is_jbod, msg="is_jbod is true")

    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    @mock.patch.object(dracclient.resources.raid.RAIDManagement,
                       'convert_physical_disks',
                       return_value={}, spec_set=True,
                       autospec=True)
    def test_raid_controller_jbod_non_raid(self, mock_requests,
                                           mock_wait_until_idrac_is_ready,
                                           mock_convert_physical_disks):

        pdv = test_utils.RAIDEnumerations[uris.DCIM_PhysicalDiskView]['ok']
        # change to non-RAID value
        pdv = pdv.replace("<n1:RaidStatus>1</n1:RaidStatus>",
                          "<n1:RaidStatus>8</n1:RaidStatus>")

        mock_requests.post(
            'https://1.2.3.4:443/wsman',
            text=pdv)

        is_jbod = self.drac_client.is_jbod_capable(self.raid_controller_fqdd)

        self.assertTrue(is_jbod, msg="is_jbod is true")

    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    @mock.patch.object(dracclient.resources.raid.RAIDManagement,
                       'convert_physical_disks',
                       return_value={}, spec_set=True,
                       autospec=True)
    def test_raid_controller_jbod_unknown(self, mock_requests,
                                          mock_wait_until_idrac_is_ready,
                                          mock_convert_physical_disks):

        is_jbod = False
        pdv = test_utils.RAIDEnumerations[uris.DCIM_PhysicalDiskView]['ok']
        # change to non-RAID value
        pdv = pdv.replace("<n1:RaidStatus>1</n1:RaidStatus>",
                          "<n1:RaidStatus>0</n1:RaidStatus>")

        mock_requests.post(
            'https://1.2.3.4:443/wsman',
            text=pdv)
        self.assertRaises(exceptions.DRACRequestFailed,
                          self.drac_client.is_jbod_capable,
                          self.raid_controller_fqdd)
        self.assertFalse(is_jbod, msg="is_jbod is false")

    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    @mock.patch.object(dracclient.resources.raid.RAIDManagement,
                       'convert_physical_disks',
                       spec_set=True,
                       autospec=True)
    def test_raid_controller_jbod_not_supported(self,
                                                mock_requests,
                                                mock_convert_physical_disks,
                                                mock_wait_idrac_is_ready):

        msg = " operation is not supported on th"
        exc = exceptions.DRACOperationFailed(drac_messages=msg)
        mock_convert_physical_disks.side_effect = exc

        mock_requests.post(
            'https://1.2.3.4:443/wsman',
            text=test_utils.RAIDEnumerations[uris.DCIM_PhysicalDiskView]['ok'])

        is_jbod = self.drac_client.is_jbod_capable(self.raid_controller_fqdd)
        self.assertFalse(is_jbod, msg="is_jbod is false")

    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    @mock.patch.object(dracclient.resources.raid.RAIDManagement,
                       'convert_physical_disks',
                       spec_set=True,
                       autospec=True)
    def test_raid_controller_jbod_ex_no_match(self,
                                              mock_requests,
                                              mock_convert_physical_disks,
                                              mock_wait_until_idrac_is_ready):

        mock_requests.post(
            'https://1.2.3.4:443/wsman',
            text=test_utils.RAIDEnumerations[uris.DCIM_PhysicalDiskView]['ok'])
        msg = "NON_MATCHING_MESSAGE"
        exc = exceptions.DRACOperationFailed(drac_messages=msg)
        mock_convert_physical_disks.side_effect = exc

        self.assertRaises(
            exceptions.DRACOperationFailed,
            self.drac_client.is_jbod_capable, self.raid_controller_fqdd)

    def test_is_raid_controller_raid(self, mock_requests):
        self.assertTrue(self.drac_client
                        .is_raid_controller("RAID.Integrated.1-1"))

    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    def test_is_raid_controller_boss(self, mock_requests,
                                     mock_wait_until_idrac_is_ready):
        mock_requests.post(
            'https://1.2.3.4:443/wsman',
            text=test_utils.RAIDEnumerations[uris.DCIM_ControllerView]['ok'])
        self.assertTrue(self.drac_client
                        .is_raid_controller("AHCI.Integrated.1-1"))

    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    def test_is_raid_controller_fail(self, mock_requests,
                                     mock_wait_until_idrac_is_ready):
        mock_requests.post(
            'https://1.2.3.4:443/wsman',
            text=test_utils.RAIDEnumerations[uris.DCIM_ControllerView]['ok'])
        self.assertFalse(self.drac_client
                         .is_raid_controller("notRAID.Integrated.1-1"))

    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    def test_is_boss_controller(self, mock_requests,
                                mock_wait_until_idrac_is_ready):
        mock_requests.post(
            'https://1.2.3.4:443/wsman',
            text=test_utils.RAIDEnumerations[uris.DCIM_ControllerView]['ok'])
        self.assertTrue(self.drac_client
                        .is_boss_controller("AHCI.Integrated.1-1"))

    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    def test_is_not_boss_controller(self, mock_requests,
                                    mock_wait_until_idrac_is_ready):
        mock_requests.post(
            'https://1.2.3.4:443/wsman',
            text=test_utils.RAIDEnumerations[uris.DCIM_ControllerView]['ok'])
        self.assertFalse(self.drac_client
                         .is_boss_controller("notAHCI.Integrated.1-1"),
                         None)

    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    def test_is_boss_controller_with_cntl_list(self, mock_requests,
                                               mock_wait_until_idrac_is_ready):
        mock_requests.post(
            'https://1.2.3.4:443/wsman',
            text=test_utils.RAIDEnumerations[uris.DCIM_ControllerView]['ok'])
        controllers = self.drac_client.list_raid_controllers()
        self.assertTrue(self.drac_client
                        .is_boss_controller("AHCI.Integrated.1-1",
                                            controllers))

    def test_check_disks_status_no_controllers(self, mock_requests):
        physical_disks = [self.disk_1, self.disk_2, self.disk_3, self.disk_4]
        raid_mgt = self.drac_client._raid_mgmt

        cont_to_phys_disk_ids = collections.defaultdict(list)
        mode = constants.RaidStatus.jbod

        raid_mgt._check_disks_status(mode, physical_disks,
                                     cont_to_phys_disk_ids)
        jbod_len = len(cont_to_phys_disk_ids['RAID.Integrated.1-1'])
        self.assertEqual(jbod_len, 0)

        # Switch mode to RAID and try again
        cont_to_phys_disk_ids = collections.defaultdict(list)
        mode = constants.RaidStatus.raid
        raid_mgt._check_disks_status(mode, physical_disks,
                                     cont_to_phys_disk_ids)
        raid_len = len(cont_to_phys_disk_ids['RAID.Integrated.1-1'])
        self.assertEqual(raid_len, 0)

    def test_check_disks_status_bad(self, mock_requests):
        mode = constants.RaidStatus.raid
        disk_2 = self.disk_2._replace(raid_status='FAKE_STATUS')
        physical_disks = [self.disk_1, disk_2, self.disk_3, self.disk_4]
        raid_mgt = self.drac_client._raid_mgmt

        self.assertRaises(ValueError,
                          raid_mgt._check_disks_status,
                          mode,
                          physical_disks,
                          self.controllers_to_physical_disk_ids)
        mode = constants.RaidStatus.jbod
        self.assertRaises(ValueError,
                          raid_mgt._check_disks_status,
                          mode,
                          physical_disks,
                          self.controllers_to_physical_disk_ids)

    def test_check_disks_status_fail(self, mock_requests):
        mode = constants.RaidStatus.raid
        disk_2_failed = self.disk_2._replace(raid_status='failed')
        physical_disks = [self.disk_1, disk_2_failed, self.disk_3, self.disk_4]
        raid_mgt = self.drac_client._raid_mgmt

        self.assertRaises(ValueError,
                          raid_mgt._check_disks_status,
                          mode,
                          physical_disks,
                          self.controllers_to_physical_disk_ids)
        mode = constants.RaidStatus.jbod
        self.assertRaises(ValueError,
                          raid_mgt._check_disks_status,
                          mode,
                          physical_disks,
                          self.controllers_to_physical_disk_ids)

    def test_check_disks_status_no_change(self, mock_requests):
        raid_mgt = self.drac_client._raid_mgmt
        mode = constants.RaidStatus.raid
        physical_disks = [self.disk_1, self.disk_2,
                          self.disk_3, self.disk_4]

        raid_cntl_to_phys_disk_ids = raid_mgt._check_disks_status(
            mode,  physical_disks, self.controllers_to_physical_disk_ids)
        raid_len = len(raid_cntl_to_phys_disk_ids['RAID.Integrated.1-1'])
        self.assertEqual(raid_len, 0)

        mode = constants.RaidStatus.jbod
        disk_1_non_raid = self.disk_1._replace(raid_status='non-RAID')
        disk_2_non_raid = self.disk_2._replace(raid_status='non-RAID')
        physical_disks = [disk_1_non_raid, disk_2_non_raid,
                          self.disk_3, self.disk_4]

        jbod_cntl_to_phys_disk_ids = raid_mgt._check_disks_status(
            mode,  physical_disks, self.controllers_to_physical_disk_ids)
        jbod_len = len(jbod_cntl_to_phys_disk_ids['RAID.Integrated.1-1'])
        self.assertEqual(jbod_len, 0)

    def test_check_disks_status_change_state(self, mock_requests):
        raid_mgt = self.drac_client._raid_mgmt
        mode = constants.RaidStatus.jbod
        physical_disks = [self.disk_1, self.disk_2, self.disk_3, self.disk_4]

        jbod_cntl_to_phys_disk_ids = raid_mgt._check_disks_status(
            mode, physical_disks, self.controllers_to_physical_disk_ids)
        jbod_len = len(jbod_cntl_to_phys_disk_ids['RAID.Integrated.1-1'])
        self.assertEqual(jbod_len, 2)

        mode = constants.RaidStatus.raid
        disk_1_non_raid = self.disk_1._replace(raid_status='non-RAID')
        disk_2_non_raid = self.disk_2._replace(raid_status='non-RAID')
        physical_disks = [disk_1_non_raid, disk_2_non_raid,
                          self.disk_3, self.disk_4]
        raid_cntl_to_phys_disk_ids = raid_mgt._check_disks_status(
            mode,  physical_disks, self.controllers_to_physical_disk_ids)
        raid_len = len(raid_cntl_to_phys_disk_ids['RAID.Integrated.1-1'])
        self.assertEqual(raid_len, 2)

    def test_check_disks_status_bad_and_fail(self, mock_requests):
        mode = constants.RaidStatus.raid
        disk_1_bad = self.disk_1._replace(raid_status='FAKE_STATUS')
        disk_2_failed = self.disk_2._replace(raid_status='failed')
        physical_disks = [disk_1_bad, disk_2_failed, self.disk_3, self.disk_4]
        raid_mgt = self.drac_client._raid_mgmt

        self.assertRaises(ValueError,
                          raid_mgt._check_disks_status,
                          mode,
                          physical_disks,
                          self.controllers_to_physical_disk_ids)
        mode = constants.RaidStatus.jbod
        self.assertRaises(ValueError,
                          raid_mgt._check_disks_status,
                          mode,
                          physical_disks,
                          self.controllers_to_physical_disk_ids)

    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    @mock.patch.object(dracclient.resources.raid.RAIDManagement,
                       'convert_physical_disks', spec_set=True,
                       autospec=True)
    def test_change_physical_disk_state_jbod(
            self, mock_requests,
            mock_convert_physical_disks,
            wait_until_idrac_is_ready):
        mode = constants.RaidStatus.jbod
        mock_requests.post(
            'https://1.2.3.4:443/wsman',
            text=test_utils.RAIDEnumerations[uris.DCIM_PhysicalDiskView]['ok'])
        cvt_phys_disks_return_value = {'is_commit_required': True,
                                       'is_reboot_required': constants.
                                       RebootRequired.true}
        mock_convert_physical_disks.return_value = cvt_phys_disks_return_value

        expected_return_value = {'RAID.Integrated.1-1':
                                 cvt_phys_disks_return_value,
                                 'AHCI.Integrated.1-1':
                                 cvt_phys_disks_return_value}
        results = self.drac_client.change_physical_disk_state(
            mode, self.controllers_to_physical_disk_ids)
        self.assertEqual(results['conversion_results'],
                         expected_return_value)

    @mock.patch.object(dracclient.resources.raid.RAIDManagement,
                       'list_physical_disks', spec_set=True,
                       autospec=True)
    @mock.patch.object(dracclient.resources.raid.RAIDManagement,
                       'convert_physical_disks', spec_set=True,
                       autospec=True)
    def test_change_physical_disk_state_raid(
            self, mock_requests,
            mock_convert_physical_disks,
            mock_list_physical_disks):
        mode = constants.RaidStatus.raid
        disk_1_non_raid = self.disk_1._replace(raid_status='non-RAID')
        disk_2_non_raid = self.disk_2._replace(raid_status='non-RAID')
        physical_disks = [disk_1_non_raid, disk_2_non_raid,
                          self.disk_3, self.disk_4]
        mock_list_physical_disks.return_value = physical_disks
        boss_return_value = {'is_commit_required': False,
                             'is_reboot_required':
                             constants.RebootRequired.false}
        raid_return_value = {'is_commit_required': True,
                             'is_reboot_required':
                             constants.RebootRequired.true}
        mock_convert_physical_disks.return_value = raid_return_value

        results = self.drac_client.change_physical_disk_state(
            mode, self.controllers_to_physical_disk_ids)
        self.assertEqual(len(results['conversion_results']), 2)
        self.assertEqual(results['conversion_results']['AHCI.Integrated.1-1'],
                         boss_return_value)
        self.assertEqual(results['conversion_results']['RAID.Integrated.1-1'],
                         raid_return_value)

    @mock.patch.object(dracclient.resources.raid.RAIDManagement,
                       'list_physical_disks', spec_set=True,
                       autospec=True)
    def test_change_physical_disk_state_none(
            self, mock_requests,
            mock_list_physical_disks):
        mode = constants.RaidStatus.raid
        physical_disks = [self.disk_1, self.disk_2, self.disk_3, self.disk_4]
        mock_list_physical_disks.return_value = physical_disks
        expected_return_value = {'is_commit_required': False,
                                 'is_reboot_required':
                                 constants.RebootRequired.false}
        results = self.drac_client.change_physical_disk_state(
            mode, self.controllers_to_physical_disk_ids)
        self.assertEqual(results['conversion_results']['RAID.Integrated.1-1'],
                         expected_return_value)
        self.assertEqual(results['conversion_results']['AHCI.Integrated.1-1'],
                         expected_return_value)

    @mock.patch.object(dracclient.resources.raid.RAIDManagement,
                       'list_physical_disks', spec_set=True,
                       autospec=True)
    @mock.patch.object(dracclient.resources.raid.RAIDManagement,
                       'convert_physical_disks', spec_set=True,
                       autospec=True,
                       side_effect=exceptions.DRACOperationFailed(
                           drac_messages=constants.NOT_SUPPORTED_MSG))
    def test_change_physical_disk_state_not_supported(
            self, mock_requests,
            mock_convert_physical_disks,
            mock_list_physical_disks):
        mode = constants.RaidStatus.raid
        disk_1_non_raid = self.disk_1._replace(raid_status='non-RAID')
        disk_2_non_raid = self.disk_2._replace(raid_status='non-RAID')
        physical_disks = [disk_1_non_raid, disk_2_non_raid,
                          self.disk_3, self.disk_4]
        mock_list_physical_disks.return_value = physical_disks
        expected_return_value = {'is_commit_required': False,
                                 'is_reboot_required':
                                 constants.RebootRequired.false}
        results = self.drac_client.change_physical_disk_state(
            mode, self.controllers_to_physical_disk_ids)
        self.assertEqual(results['conversion_results']['RAID.Integrated.1-1'],
                         expected_return_value)
        self.assertEqual(results['conversion_results']['AHCI.Integrated.1-1'],
                         expected_return_value)

    @mock.patch.object(dracclient.resources.raid.RAIDManagement,
                       'list_physical_disks', spec_set=True,
                       autospec=True)
    @mock.patch.object(dracclient.resources.raid.RAIDManagement,
                       'convert_physical_disks', spec_set=True,
                       autospec=True,
                       side_effect=exceptions.DRACOperationFailed(
                           drac_messages="OTHER_MESSAGE"))
    def test_change_physical_disk_state_raise_drac_operation_other(
            self, mock_requests,
            mock_convert_physical_disks,
            mock_list_physical_disks):
        mode = constants.RaidStatus.raid
        disk_1_non_raid = self.disk_1._replace(raid_status='non-RAID')
        disk_2_non_raid = self.disk_2._replace(raid_status='non-RAID')
        physical_disks = [disk_1_non_raid, disk_2_non_raid,
                          self.disk_3, self.disk_4]
        mock_list_physical_disks.return_value = physical_disks
        self.assertRaisesRegexp(
            exceptions.DRACOperationFailed,
            "OTHER_MESSAGE",
            self.drac_client.change_physical_disk_state,
            mode,
            self.controllers_to_physical_disk_ids)

    @mock.patch.object(dracclient.resources.raid.RAIDManagement,
                       'list_physical_disks', spec_set=True,
                       autospec=True)
    @mock.patch.object(dracclient.resources.raid.RAIDManagement,
                       'convert_physical_disks', spec_set=True,
                       autospec=True, side_effect=Exception(
                           "SOMETHING_BAD_HAPPENED"))
    def test_change_physical_disk_state_raise_other(
            self, mock_requests,
            mock_convert_physical_disks,
            mock_list_physical_disks):
        mode = constants.RaidStatus.raid
        disk_1_non_raid = self.disk_1._replace(raid_status='non-RAID')
        disk_2_non_raid = self.disk_2._replace(raid_status='non-RAID')
        physical_disks = [disk_1_non_raid, disk_2_non_raid,
                          self.disk_3, self.disk_4]
        mock_list_physical_disks.return_value = physical_disks
        self.assertRaisesRegexp(
            Exception,
            "SOMETHING_BAD_HAPPENED",
            self.drac_client.change_physical_disk_state,
            mode,
            self.controllers_to_physical_disk_ids)

    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    @mock.patch.object(dracclient.resources.raid.RAIDManagement,
                       'list_physical_disks', spec_set=True,
                       autospec=True)
    @mock.patch.object(dracclient.resources.raid.RAIDManagement,
                       'convert_physical_disks', spec_set=True,
                       autospec=True)
    def test_change_physical_disk_state_with_no_dict(
            self, mock_requests,
            mock_convert_physical_disks,
            mock_list_physical_disks,
            mock_wait_until_idrac_is_ready):
        mock_requests.post(
            'https://1.2.3.4:443/wsman',
            text=test_utils.RAIDEnumerations[uris.DCIM_ControllerView]['ok'])
        mode = constants.RaidStatus.jbod
        physical_disks = [self.disk_1, self.disk_2, self.disk_3, self.disk_4]
        mock_list_physical_disks.return_value = physical_disks
        cvt_phys_disks_return_value = {'is_commit_required': True,
                                       'is_reboot_required': constants.
                                       RebootRequired.true}
        mock_convert_physical_disks.return_value = cvt_phys_disks_return_value
        expected_return_value = {'RAID.Integrated.1-1':
                                 cvt_phys_disks_return_value,
                                 'AHCI.Integrated.1-1':
                                 cvt_phys_disks_return_value}
        results = self.drac_client.change_physical_disk_state(mode)
        self.assertDictEqual(results['conversion_results'],
                             expected_return_value)

    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    @mock.patch.object(dracclient.resources.raid.RAIDManagement,
                       'list_physical_disks', spec_set=True,
                       autospec=True)
    def test_change_physical_disk_state_with_no_raid_or_boss_card_match(
            self, mock_requests,
            mock_list_physical_disks,
            mock_wait_until_idrac_is_ready):
        mock_requests.post(
            'https://1.2.3.4:443/wsman',
            text=test_utils.RAIDEnumerations[uris.DCIM_ControllerView]['ok'])
        mode = constants.RaidStatus.jbod
        _disk_1 = self.disk_1._replace(controller='NOT_RAID.Integrated.1-1')
        _disk_2 = self.disk_2._replace(controller='NOT_RAID.Integrated.1-1')
        _disk_3 = self.disk_3._replace(controller='NOT_AHCI.Integrated.1-1')
        _disk_4 = self.disk_4._replace(controller='NOT_AHCI.Integrated.1-1')
        physical_disks = [_disk_1, _disk_2, _disk_3, _disk_4]
        mock_list_physical_disks.return_value = physical_disks
        results = self.drac_client.change_physical_disk_state(mode)
        self.assertDictEqual(results['conversion_results'], {})

    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    def test_is_realtime_supported_with_realtime_controller(
            self,
            mock_requests,
            mock_wait_until_idrac_is_ready):
        expected_raid_controller = 'RAID.Integrated.1-1'

        mock_requests.post(
            'https://1.2.3.4:443/wsman',
            text=test_utils.RAIDEnumerations[uris.DCIM_ControllerView]['ok'])

        self.assertTrue(
                self.drac_client.is_realtime_supported(
                    expected_raid_controller))

    @mock.patch.object(dracclient.client.WSManClient,
                       'wait_until_idrac_is_ready', spec_set=True,
                       autospec=True)
    def test_is_realtime_supported_with_non_realtime_controller(
            self,
            mock_requests,
            mock_wait_until_idrac_is_ready):
        expected_raid_controller = 'AHCI.Integrated.1-1'

        mock_requests.post(
            'https://1.2.3.4:443/wsman',
            text=test_utils.RAIDEnumerations[uris.DCIM_ControllerView]['ok'])

        self.assertFalse(
                self.drac_client.is_realtime_supported(
                    expected_raid_controller))
